#include "spheroidal.h"

//  -----------------------------------------------
void coord_st_To_xy(parameters par, int idom, int i, int j, derivs_1D G, coordinate *x, coordinate *y)
{// Calculation of the coordinates (x,y) from (s,t) along with first and second
 // derivatives with repect to s and t
	switch(idom){
		case 0: coord_0(par, i, j, G, x, y); break;
		case 1: coord_1(par, i, j, G, x, y); break;
	}
}
//  -----------------------------------------------
void Dcoord_st_To_xy(parameters par, parameters Dpar, int idom, int i, int j, derivs_1D G, derivs_1D DG, 
	coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy)
{// Linear version of 'coord_st_To_xy'
	switch(idom){
		case 0: Dcoord_0(par, Dpar, i, j, G, DG, x, y, Dx, Dy); break;
		case 1: Dcoord_1(par, Dpar, i, j, G, DG, x, y, Dx, Dy); break;
	}
}
//   -----------------------------------------------
void get_sig(parameters par, int idom, int i, double *sig, double *sig_m_1, double *sig_s, double *sig_ss)
{
	int ns=par.ns[idom], Ns=ns-1;
	double del_s = par.del_s, aux = 1./sinch(del_s),
		Pih_m_a = Pih*(Ns-i)/Ns, ca=sin(Pih_m_a), sm1=-ca*ca; // Pih_m_a = Pi/2-a

		*sig_m_1 = sm1*sinch(del_s*sm1)*aux;
		*sig     = 1. + (*sig_m_1);
		*sig_s   = cosh(del_s*sm1)*aux;
		*sig_ss  = sinh(del_s*sm1)*del_s*aux;
}
//   -----------------------------------------------
void get_tau(parameters par, int idom, int j, double *tau, double *tau_m_1, double *tau_t, double *tau_tt)
{
	int nt=par.nt[idom], Nt=nt-1;
	double del_t = par.del_t, aux = 1./sinch(del_t),
		b = Pih*j/Nt, sb=sin(b), t=sb*sb;

		*tau     = t*    sinch(del_t*t)*aux;     // tau    = sinh(del*t)/sinh(del)
		*tau_m_1 = *tau - 1.;
		*tau_t   =       cosh( del_t*t)*aux;     // tau_t  = del*cosh(del*t)/sinh(del)
		*tau_tt  = del_t*sinh( del_t*t)*aux;     // tau_tt = del^2*sinh(del*t)/sinh(del)
}
//   -----------------------------------------------
/*void get_tau(parameters par, int idom, int j, double *tau, double *tau_m_1, double *tau_t, double *tau_tt)
{
	int nt=par.nt[idom], Nt=nt-1;
	double del_t = par.del_t, aux = 1./sinch(del_t),
		Pih_m_b = Pih*(Nt-j)/Nt, cb=sin(Pih_m_b), tm1=-cb*cb; // Pih_m_b = Pi/2-b

		*tau_m_1 = tm1*sinch(del_t*tm1)*aux;
		*tau     = 1. + (*tau_m_1);
		*tau_t   = cosh(del_t*tm1)*aux;
		*tau_tt  = sinh(del_t*tm1)*del_t*aux;
}*/
//   -----------------------------------------------
void coord_0(parameters par, int i, int j, derivs_1D G, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 0
	int  idom = 0, nt=par.nt[idom], Nt=nt-1;
	double sig, sig_m_1, sig2, sig_s, sig2_s, sig_ss, tau, tau_m_1, tau_t, tau_tt, 
		xi, xi_s, xi_ss, xi2, xi2_s, xi2_ss,
		yp = G.d0[0],  rp = sqrt(yp),
		xe = G.d0[Nt], re = sqrt(xe);

	get_sig(par, idom, i, &sig, &sig_m_1, &sig_s, &sig_ss);
	get_tau(par, idom, j, &tau, &tau_m_1, &tau_t, &tau_tt);

	sig2    = sig*sig;
	sig2_s  = 2.*sig*sig_s;
	xi      = rp - re*sig_m_1/sig;
	xi_s    = -re*sig_s/sig2;
	xi_ss   = -re*sig_ss/sig2 - xi_s*sig2_s/sig2;
	xi2     = xi*xi;
	xi2_s   = 2.*xi*xi_s;
	xi2_ss  = 2.*(xi*xi_ss + xi_s*xi_s);

	(*x).d0  = (xi2 + xe - yp)*tau;
	(*x).d1  = xi2_s*tau;
	(*x).d11 = xi2_ss*tau;
	(*x).d2  = (xi2 + xe - yp)*tau_t;
	(*x).d22 = (xi2 + xe - yp)*tau_tt;
	(*x).d12 = xi2_s*tau_t;

	(*y).d0  =  fabs(tau_m_1*(yp-xi2) - xe*tau + G.d0[j]);
	(*y).d1  = -tau_m_1*xi2_s;
	(*y).d11 = -tau_m_1*xi2_ss;
	(*y).d2  =  tau_t* (yp-xi2-xe) +  G.d2[j];
	(*y).d22 =  tau_tt*(yp-xi2-xe) + G.d22[j];
	(*y).d12 = -tau_t*xi2_s;
}
//   -----------------------------------------------
void Dcoord_0(parameters par, parameters Dpar, int i, int j, derivs_1D G, derivs_1D DG, 
	coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy)
{// // Linear version of 'coord_0'
	int  idom = 0, nt=par.nt[idom], Nt=nt-1;
	double sig, sig_m_1, sig2, sig_s, sig2_s, sig_ss, tau, tau_m_1, tau_t, tau_tt, 
		xi, xi_s, xi_ss, xi2, xi2_s, xi2_ss, Dxi, Dxi_s, Dxi_ss, Dxi2, Dxi2_s, Dxi2_ss,
		yp = G.d0[0],  rp = sqrt(yp), Dyp = DG.d0[0],  Drp = 0.5*Dyp/rp,
		xe = G.d0[Nt], re = sqrt(xe), Dxe = DG.d0[Nt], Dre = 0.5*Dxe/re;

	get_sig(par, idom, i, &sig, &sig_m_1, &sig_s, &sig_ss);
	get_tau(par, idom, j, &tau, &tau_m_1, &tau_t, &tau_tt);

	sig2    = sig*sig;
	sig2_s  = 2.*sig*sig_s;
	xi      = rp - re*sig_m_1/sig;                        Dxi     =  Drp - Dre*sig_m_1/sig;
	xi_s    = -re*sig_s/sig2;                             Dxi_s   = -Dre*sig_s/sig2;
	xi_ss   = -re*sig_ss/sig2 - xi_s*sig2_s/sig2;         Dxi_ss  = -Dre*sig_ss/sig2 - Dxi_s*sig2_s/sig2;
	xi2     = xi*xi;                                      Dxi2    = 2.*xi*Dxi,
	xi2_s   = 2.*xi*xi_s;                                 Dxi2_s  = 2.*(Dxi*xi_s  + xi*Dxi_s),
	xi2_ss  = 2.*(xi*xi_ss + xi_s*xi_s);                  Dxi2_ss = 2.*(Dxi*xi_ss + xi*Dxi_ss + 2.*Dxi_s*xi_s);

	(*x).d0  = (xi2 + xe - yp)*tau;                       (*Dx).d0  = (Dxi2 + Dxe - Dyp)*tau;
	(*x).d1  = xi2_s*tau;                                 (*Dx).d1  = Dxi2_s*tau;
	(*x).d11 = xi2_ss*tau;                                (*Dx).d11 = Dxi2_ss*tau;
	(*x).d2  = (xi2 + xe - yp)*tau_t;                     (*Dx).d2  = (Dxi2 + Dxe - Dyp)*tau_t;
	(*x).d22 = (xi2 + xe - yp)*tau_tt;                    (*Dx).d22 = (Dxi2 + Dxe - Dyp)*tau_tt;
	(*x).d12 = xi2_s*tau_t;                               (*Dx).d12 = Dxi2_s*tau_t;

	(*y).d0  =  fabs(tau_m_1*(yp-xi2)-xe*tau+G.d0[j]);    (*Dy).d0  = tau_m_1*(Dyp-Dxi2)-Dxe*tau+DG.d0[j];
	(*y).d1  = -tau_m_1*xi2_s;                            (*Dy).d1  = -tau_m_1*Dxi2_s;
	(*y).d11 = -tau_m_1*xi2_ss;                           (*Dy).d11 = -tau_m_1*Dxi2_ss;
	(*y).d2  =  tau_t* (yp-xi2-xe) +  G.d2[j];            (*Dy).d2  = tau_t* (Dyp-Dxi2-Dxe) +  DG.d2[j];
	(*y).d22 =  tau_tt*(yp-xi2-xe) + G.d22[j];            (*Dy).d22 = tau_tt*(Dyp-Dxi2-Dxe) + DG.d22[j];
	(*y).d12 = -tau_t*xi2_s;                              (*Dy).d12 = -tau_t*Dxi2_s;
}
//   -----------------------------------------------
void coord_1(parameters par, int i, int j, derivs_1D G, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 1
	int idom = 1, ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1;
	double a = Pih*i/Ns, sa=sin(a), s=sa*sa, tau, tau_m_1, tau_t, tau_tt, xe = G.d0[Nt];

	get_tau(par, idom, j, &tau, &tau_m_1, &tau_t, &tau_tt);

	(*x).d0   = xe*tau;
	(*x).d2   = xe*tau_t;
	(*x).d22  = xe*tau_tt;
	(*x).d1   = (*x).d11 = (*x).d12 = 0.;
	
	(*y).d11  = 0.;
	(*y).d1   = fabs(G.d0[j]  - xe*tau);
	(*y).d12  =      G.d2[j]  - xe*tau_t;
	(*y).d122 =      G.d22[j] - xe*tau_tt;
	(*y).d0   = s*(*y).d1;
	(*y).d2   = s*(*y).d12;
	(*y).d22  = s*(*y).d122;
}
//---------------------------------------------------------------
void Dcoord_1(parameters par, parameters Dpar, int i, int j, derivs_1D G, derivs_1D DG, 
	coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy)
{// Linear version of 'coord_1'
	int idom = 1, ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1;
	double a = Pih*i/Ns, sa=sin(a), s=sa*sa, tau, tau_m_1, tau_t, tau_tt, xe = G.d0[Nt], Dxe = DG.d0[Nt];

	get_tau(par, idom, j, &tau, &tau_m_1, &tau_t, &tau_tt);

	(*x).d0   = xe*tau;                    (*Dx).d0   = Dxe*tau;
	(*x).d2   = xe*tau_t;                  (*Dx).d2   = Dxe*tau_t;
	(*x).d22  = xe*tau_tt;                 (*Dx).d22  = Dxe*tau_tt;
	(*x).d1   = (*x).d11 = (*x).d12 = 0.;  (*Dx).d1   = (*Dx).d11 = (*Dx).d12 = 0.;
	
	(*y).d11  = 0.;                        (*Dy).d11  = 0.;
	(*y).d1   = fabs(G.d0[j] - xe*tau);    (*Dy).d1   = DG.d0[j] - Dxe*tau;
	(*y).d12  =      G.d2[j] - xe*tau_t;   (*Dy).d12  = DG.d2[j] - Dxe*tau_t;
	(*y).d122 =      G.d22[j]- xe*tau_tt;  (*Dy).d122 = DG.d22[j]- Dxe*tau_tt;
	(*y).d0   = s*(*y).d1;                 (*Dy).d0   = s*(*Dy).d1;
	(*y).d2   = s*(*y).d12;                (*Dy).d2   = s*(*Dy).d12;
	(*y).d22  = s*(*y).d122;               (*Dy).d22  = s*(*Dy).d122;
}
//  -----------------------------------------------
