#include "spheroidal.h"

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
// The Calculation of F(X) and J*DX, as explained on pages 12-14
// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------


// -------------------------------------------------------------------------------
void v_To_V(parameters par, int idom, int i, int j, derivs_2D v, derivs_1D w, derivs_2D V)
{	// Writes (v) at gridpoint (i,j) in domain (idom) to the auxiliary derivs (V)
	int ipot;

	if(idom == NDOM-1){
		for(ipot=0; ipot<NPOT; ipot++){
			int indx = Index(par,idom,ipot,i,j), indx_S = Index_Star(par,ipot,j);
			V.d0[ipot]  = w.d0[indx_S]  + v.d0[indx];
			V.d1[ipot]  =                 v.d1[indx];
			V.d2[ipot]  = w.d2[indx_S]  + v.d2[indx];
			V.d11[ipot] =                 v.d11[indx];
			V.d22[ipot] = w.d22[indx_S] + v.d22[indx];
			V.d12[ipot] =                 v.d12[indx];
		}
	}	
	else{
		for(ipot=0; ipot<NPOT; ipot++){
			int indx = Index(par, idom,ipot,i,j);
			V.d0[ipot]  = v.d0[indx];  // V
			V.d1[ipot]  = v.d1[indx];  // V_s
			V.d2[ipot]  = v.d2[indx];  // V_t
			V.d11[ipot] = v.d11[indx]; // V_ss
			V.d12[ipot] = v.d12[indx]; // V_st
			V.d22[ipot] = v.d22[indx]; // V_tt
		}
	}
}
// -------------------------------------------------------------------------------
void Derivatives_st(parameters par, derivs_2D v)
{
	int idom, ipot, ns, nt, i, j, indx[NMAX];
	double p[NMAX], dp[NMAX], d2p[NMAX], q[NMAX], dq[NMAX];
	
	for(idom=0; idom<NDOM; idom++){

		ns=par.ns[idom];
		nt=par.nt[idom];

		for(ipot=0; ipot<NPOT; ipot++){

			for(j=0;j<nt;j++) {		// Calculation of Derivatives w.r.t. s-Direction
				for(i=0;i<ns;i++){	// (Chebyshev_Extremes)
					indx[i] = Index(par,idom,ipot,i,j);
					p[i]    = v.d0[indx[i]];
				}
				chebft_Extremes(p,ns,0);
				chder(0., 1., p,dp,ns);
				chder(0., 1., dp,d2p,ns);
				chebft_Extremes(dp,ns,1);
				chebft_Extremes(d2p,ns,1);
				for(i=0;i<ns;i++){
					v.d1[indx[i]]  = dp[i];
					v.d11[indx[i]] = d2p[i];
				}
			}  

			for(i=0;i<ns;i++) {		// Calculation of Derivatives w.r.t. t-Direction
				for(j=0;j<nt;j++){	// (Chebyshev_Extremes)
					indx[j] = Index(par,idom,ipot,i,j);
					p[j]    = v.d0[indx[j]];
					q[j]    = v.d1[indx[j]];
				}
				chebft_Extremes(p,nt,0);
				chebft_Extremes(q,nt,0);
				chder(0., 1., p,dp,nt);
				chder(0., 1., dp,d2p,nt);
				chder(0., 1., q,dq,nt);
				chebft_Extremes(dp,nt,1);
				chebft_Extremes(d2p,nt,1);
				chebft_Extremes(dq,nt,1);
				for(j=0;j<nt;j++){
					v.d2[indx[j]]  = dp[j];
					v.d22[indx[j]] = d2p[j];
					v.d12[indx[j]] = dq[j];
				}
			}

		}
	}

}
// -------------------------------------------------------------------------------
void Derivatives_t(parameters par, derivs_1D w, derivs_1D G)
{
	int nt=par.nt[NDOM-1], ipot, j, indx[NMAX];
	double p[NMAX], dp[NMAX], d2p[NMAX];
	
	for(ipot=0; ipot<=NPOT; ipot++){
		for(j=0;j<nt;j++){	// Calculation of Derivatives w.r.t. t-Direction (Chebyshev_Extremes)
			if (ipot < NPOT) {
				indx[j] = Index_Star(par,ipot,j);
				p[j]    = w.d0[indx[j]];
			}
			else
				p[j]    = G.d0[j];
		}
		chebft_Extremes(p,nt,0);
		chder(0., 1., p,dp,nt);
		chder(0., 1., dp,d2p,nt);
		chebft_Extremes(dp,nt,1);
		chebft_Extremes(d2p,nt,1);
		for(j=0;j<nt;j++){
			if (ipot < NPOT) {
				w.d2[indx[j]]  = dp[j];
				w.d22[indx[j]] = d2p[j];
			}
			else{
				G.d2[j]        = dp[j];
				G.d22[j]       = d2p[j];
			}
		}
	}  
}
//  -----------------------------------------------
void Get_vwGpar_From_X(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters *parX)
{
	int n_2D=par.n_2D, n, n_v, idom, ipot, i, j, ns, Ns, nt, Nt;
	
	ns = par.ns[NDOM-1];
	Ns = ns-1;
	for(n=0; n<n_2D; n++){
		n_v  = par.n_v_of_n[n];
		Get_Indices_From_n_v(par, n_v, &idom, &ipot, &i, &j);

		if (idom < NDOM-1 || (idom == NDOM-1 && i < Ns))     v.d0[n_v]   = X[n];
		else{
			v.d0[n_v] = 0.;
			w.d0[Index_Star(par,ipot,j)] = X[n];
		}
	}
	Derivatives_st(par, v);

	nt = par.nt[NDOM-1];
	Nt = nt-1;
	for(j=0; j<nt; j++) G.d0[j] = X[n_2D + j];

	Derivatives_t(par, w, G);

	(*parX).V0      = X[n_2D + nt + 0]; // X[n_2D + nt + 0] = V0
	(*parX).Omega     = X[n_2D + nt + 1]; // X[n_2D + nt + 1] = Omega (axis value of Omega)

	(*parX).rp      = sqrt(G.d0[0]);
	(*parX).re      = sqrt(G.d0[Nt]);
	(*parX).r_ratio = (*parX).rp/(*parX).re;

	(*parX).xp      = G.d0[0];
	(*parX).xe      = G.d0[Nt];
}
// -------------------------------------------------------------------------------
void F_of_X(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters *parX, double *F)
{
	int n_2D = par.n_2D, ntotal = par.ntotal, idom, i, j, ns, Ns, nt, Nt, n, n_v;
	double *FF;
	derivs_2D V;

	allocate_derivs_2D(&V, NPOT);
	FF = dvector(0, ntotal-1);
	fill0_dvector(FF, 0, ntotal);

	*parX = par;
	Get_vwGpar_From_X(par, X, v, w, G, parX);

	for(idom=0; idom<NDOM; idom++){
		ns=par.ns[idom];		Ns=ns-1;
		nt=par.nt[idom];		Nt=nt-1;
		for(j=0;j<nt;j++) {
			for(i=0; i<ns; i++){
				if (i*j*(Ns-i)*(Nt-j) != 0)
					NonLinFieldEqns(*parX, idom, i, j, v, w, G, V, FF);
				else
					NonLinBoundEqns(*parX, idom, i, j, v, w, G, V, FF);
			}
			if(idom == NDOM-1) NonLinStarBoundary(*parX, j, w, G, V, FF);
		}
	}
	NonLinParamEqns(par, *parX, v, w, G, V, FF);

	for(n=0; n < ntotal; n++){ // reordering
		if(n < n_2D) n_v = par.n_v_of_n[n];
		else         n_v = n;
		F[n] = FF[n_v];
	}
	
	free_derivs_2D(&V, NPOT);
	free_dvector(FF, 0, ntotal-1);
}
// -------------------------------------------------------------------------------
void J_times_DX(parameters par, double *X, double *DX, derivs_2D v, derivs_1D w, derivs_1D G,
	parameters parX, double *JDX)
{
	int n_2D = par.n_2D, ntotal = par.ntotal, idom, i, j, ns, Ns, nt, Nt, n, n_v;
	parameters DparX=par;
	double *JJDX;
	derivs_1D DG, Dw;
	derivs_2D Dv, V, DV;

	nt = par.nt[NDOM-1];
	allocate_derivs_1D(&DG, nt);
	allocate_derivs_1D(&Dw, nt*NPOT);
	allocate_derivs_2D(&Dv, n_2D);
	allocate_derivs_2D( &V, NPOT);
	allocate_derivs_2D(&DV, NPOT);
	JJDX = dvector(0, ntotal-1);
	fill0_dvector(JJDX, 0, ntotal);

	Get_vwGpar_From_X(par, DX, Dv, Dw, DG, &DparX);

	for(idom=0; idom<NDOM; idom++){
		ns=par.ns[idom];		Ns=ns-1;
		nt=par.nt[idom];		Nt=nt-1;
		for(j=0;j<nt;j++) {
			for(i=0;i<ns;i++){
				if (i*j*(Ns-i)*(Nt-j) != 0)
					LinFieldEqns(parX, DparX, idom, i, j, v, w, G, V, Dv, Dw, DG, DV, JJDX);
				else
					LinBoundEqns(parX, DparX, idom, i, j, v, w, G, V, Dv, Dw, DG, DV, JJDX);
			}
			if(idom == NDOM-1) LinStarBoundary(parX, DparX, j, w, G, V, Dw, DG, DV, JJDX);
		}
	}
	LinParamEqns(par, parX, DparX, v, w, G, V, Dv, Dw, DG, DV, JJDX);

	for(n=0; n < ntotal; n++){ // reordering
		if(n < n_2D) n_v = par.n_v_of_n[n];
		else         n_v = n;
		JDX[n] = JJDX[n_v];
	}
	
	nt = par.nt[NDOM-1];
	free_derivs_1D(&DG, nt);
	free_derivs_1D(&Dw, nt*NPOT);
	free_derivs_2D(&Dv, n_2D);
	free_derivs_2D( &V, NPOT);
	free_derivs_2D(&DV, NPOT);
	free_dvector(JJDX, 0, ntotal-1);
}
