#include "spheroidal.h"

// -------------------------------------------------------------------------------
// Calculating an approximate solution corresponding to new resolution orders through
// spectral interpolation
// -------------------------------------------------------------------------------
void v1_To_v2(parameters par1, derivs_2D v1, parameters par2, derivs_2D v2)
{
	int	ipot, idom, i_1, j_1, i_2, j_2, indx;
	double p[NMAX], *v21;
	parameters par21=par1;

	par21.nst[0] = 0;
	for(idom=0; idom<NDOM; idom++){
		par21.ns[idom]    = par2.ns[idom]; 
		par21.nt[idom]    = par1.nt[idom]; 
		par21.nst[idom+1] = par21.nst[idom] + par21.ns[idom]*par21.nt[idom];
	}
	v21 = dvector(0, NPOT*par21.nst[NDOM]-1);	

	for(idom=0; idom<NDOM; idom++){
		int 
			ns_1=par1.ns[idom], nt_1=par1.nt[idom],
			ns_2=par2.ns[idom], nt_2=par2.nt[idom], Ns_2=ns_2-1, Nt_2=nt_2-1;
		for(ipot=0; ipot < NPOT; ipot++) {
			for(j_1=0; j_1<nt_1; j_1++){
				for(i_1=0; i_1<ns_1; i_1++){
					indx   = Index(par1, idom, ipot, i_1, j_1);
					p[i_1] = v1.d0[indx];
				}
				chebft_Extremes(p,ns_1,0);
				for(i_2=0; i_2<ns_2; i_2++) {
					double a=Pih*i_2/Ns_2, sa=sin(a), s=sa*sa;
					indx      = Index(par21, idom, ipot, i_2, j_1);
					v21[indx] = chebev(0., 1., p, ns_1, s);
				}
			}
			for(i_2=0; i_2<ns_2; i_2++){
				for(j_1=0; j_1<nt_1; j_1++){
					indx   = Index(par21, idom, ipot, i_2, j_1);
					p[j_1] = v21[indx];
				}
				chebft_Extremes(p,nt_1,0);
				for(j_2=0; j_2<nt_2; j_2++) {
					double b=Pih*j_2/Nt_2, sb=sin(b), t=sb*sb;
					indx        = Index(par2, idom,ipot,i_2, j_2);
					v2.d0[indx] = chebev(0., 1., p, nt_1, t);
				}
			}
		}
	}	
	free_dvector(v21, 0, NPOT*par21.nst[NDOM]-1);
}
// -------------------------------------------------------------------------------
void wG1_To_wG2(parameters par1, derivs_1D w1, derivs_1D G1, parameters par2, derivs_1D w2, derivs_1D G2) 
{
	int	idom=NDOM-1, nt_1=par1.nt[idom], nt_2=par2.nt[idom], Nt_2=nt_2-1, ipot, j_1, j_2;
	double p[NMAX];
	
	for(ipot=0; ipot <= NPOT; ipot++) {
		for(j_1=0; j_1<nt_1; j_1++){
			if (ipot < NPOT) p[j_1] = w1.d0[Index_Star(par1, ipot, j_1)];
			else             p[j_1] = G1.d0[j_1];
		}
		chebft_Extremes(p,nt_1,0);
		for(j_2=0; j_2<nt_2; j_2++) {
			double b=Pih*j_2/Nt_2, sb=sin(b), t=sb*sb,
				f=chebev(0., 1., p, nt_1, t);
			if (ipot < NPOT) w2.d0[Index_Star(par2, ipot, j_2)] = f;
			else             G2.d0[j_2]                         = f; 
		}
	}
}
//   -----------------------------------------------
void GetNewResolution(parameters par0, parameters par, double *X0, double *X)
{
	int idom, ipot, ipar, i, j, nt, ns, Ns, n, n_v, n_2D;
	derivs_2D v0, v;
	derivs_1D G0, w0, G, w;

	nt = par0.nt[NDOM-1];
	allocate_derivs_1D(&G0, nt);
	allocate_derivs_1D(&w0, nt*NPOT);
	allocate_derivs_2D(&v0, par0.n_2D);
	Get_vwGpar_From_X(par0, X0, v0, w0, G0, &par0);
	
	nt = par.nt[NDOM-1];
	allocate_derivs_1D(&G, nt);
	allocate_derivs_1D(&w, nt*NPOT);
	allocate_derivs_2D(&v, par.n_2D);
	v1_To_v2(  par0, v0,     par, v);
	wG1_To_wG2(par0, w0, G0, par, w, G);
	
	n_2D = par.n_2D;
	ns   = par.ns[NDOM-1];
	Ns   = ns-1;
	for(n=0; n<par.n_2D; n++){
		n_v  = par.n_v_of_n[n];
		Get_Indices_From_n_v(par, n_v, &idom, &ipot, &i, &j);
		
		if (idom < NDOM-1 || (idom == NDOM-1 && i < Ns)) X[n] = v.d0[n_v];
		else                                             X[n] = w.d0[Index_Star(par, ipot, j)];
	}
	nt = par.nt[NDOM-1];
	for(j=0; j<nt; j++)
		X[n_2D + j] = G.d0[j];
	
	for(ipar=0; ipar<NPAR; ipar++)
		X[n_2D + nt + ipar] = X0[par0.n_2D + par0.nt[NDOM-1] + ipar];

	nt = par.nt[NDOM-1];
	free_derivs_2D(&v, par.n_2D);
	free_derivs_1D(&w, nt*NPOT);
	free_derivs_1D(&G, nt);
	
	nt = par0.nt[NDOM-1];
	free_derivs_2D(&v0, par0.n_2D);
	free_derivs_1D(&w0, nt*NPOT);
	free_derivs_1D(&G0, nt);
}
