#include "spheroidal.h"

//--------------------------------------------------------------------------------
int GradIndex(int i, int j)
{
	return i*NPOT + j;
}
// -------------------------------------------------------------------------------
int Index_i(parameters par, int idom, int i)
{
	int Ns = par.ns[idom]-1, i1 = i;
	
	if (i1 < 0)  i1 = -i1;
	if (i1 > Ns) i1 = 2*Ns-i1;
	
	return i1;
}
// -------------------------------------------------------------------------------
int Index_j(parameters par, int idom, int j)
{
	int Nt = par.nt[idom]-1, j1 = j;
	
	if (j1 < 0)  j1 = -j1;
	if (j1 > Nt) j1 = 2*Nt-j1;
	
	return j1;
}
// -------------------------------------------------------------------------------
int Index(parameters par, int idom, int ipot, int i, int j)
{	// nst[idom] = Sum[ ns[jdom]*nt[jdom] , {jdom, 0, idom-1} ]; nst[0] = 0; see page 6
	int ns = par.ns[idom],           nst = par.nst[idom],
	    ii = Index_i(par, idom, i),   jj = Index_j(par, idom, j);

	return   ipot + NPOT*(ii + ns*jj + nst);
}
// -------------------------------------------------------------------------------
int Index_Star(parameters par, int ipot, int j)
{
	return   ipot + NPOT*Index_j(par, NDOM-1, j);
}
// -------------------------------------------------------------------------------
void Get_Arrays_n_And_n_v_domain_0(parameters *par, int *n)
{	 // Outside the Star: V(s,t) = v(s,t)
	int idom=0, ns=(*par).ns[idom], Ns=ns-1, nt=(*par).nt[idom], Nt=nt-1, i, j, ipot, n_v;

	for(i=0;i<=Ns;i++){     // Indices of v^{0;ipot}_{ij}
		for(j=0;j<=Nt;j++) {
			for(ipot=0; ipot<NPOT; ipot++){
				n_v = Index(*par, idom, ipot, i, j);
				(*par).n_v_of_n[*n]  = n_v;
				(*par).n_of_n_v[n_v] = *n;
				*n += 1;
			}
		}
	}
}
// -------------------------------------------------------------------------------
void Get_Arrays_n_And_n_v_domain_1(parameters *par, int *n)
{	// Inside the Star: V(s,t) = w(t) + v(s,t); v(1,t) = 0
	int idom=1, ns=(*par).ns[idom], Ns=ns-1, nt=(*par).nt[idom], Nt=nt-1, i, j, ipot, n_v;

	for(i=Ns-1;i>=0;i--){  // Indices of v^{1;ipot}_{ij}
		for(j=0;j<=Nt;j++) {
			for(ipot=0; ipot<NPOT; ipot++){
				n_v = Index(*par, idom, ipot, i, j);
				(*par).n_v_of_n[*n]  = n_v;
				(*par).n_of_n_v[n_v] = *n;
				*n += 1;
			}
		}
	}
	// now: *n = num_v = n_2D - nt*NPOT:
	//     in X[0]..X[num_v]: all non-trivial potentials v in domains 0 and 1
	//    [in domain 1, v has the trivial values v(s=1,t) = 0]
	(*par).num_v = (*par).n_2D - nt*NPOT; //  = *n

	i=Ns;        // Indices of w^{1;ipot}_j
	for(j=0;j<=Nt;j++) {
		for(ipot=0; ipot<NPOT; ipot++){
			n_v = Index(*par, idom, ipot, i, j);
			(*par).n_v_of_n[*n]  = n_v;
			(*par).n_of_n_v[n_v] = *n;
			*n += 1;
		}
	}
}
// -------------------------------------------------------------------------------
void Get_Arrays_n_And_n_v(parameters *par)
{
	int n = 0;

	(*par).n_2D      = NPOT*(*par).nst[NDOM]; // #(all non-trivial Potential-values v) + #(all Potential-values w)

	(*par).n_of_n_v  = ivector(0, (*par).n_2D-1);
	(*par).n_v_of_n  = ivector(0, (*par).n_2D-1);

	Get_Arrays_n_And_n_v_domain_0(par, &n);
	Get_Arrays_n_And_n_v_domain_1(par, &n);
	// ntotal = n_2D + #(values of G) + #(Parameters)
	(*par).ntotal = (*par).n_2D + (*par).nt[NDOM-1] + NPAR;
}
// -------------------------------------------------------------------------------
void Get_Indices_From_n_v(parameters par, int n_v, int *idom, int *ipot, int *i, int *j)
{
	int aux1, aux2, ns;

	*idom = NDOM;
	while(n_v < NPOT*par.nst[*idom]) *idom -= 1;

	if (*idom<NDOM){
		aux1    = n_v - NPOT*par.nst[*idom];
		ns      = par.ns[*idom];
		*j      = aux1/(NPOT*ns);
		aux2    = aux1 - NPOT*ns*(*j);
		*i      = aux2/NPOT;
		*ipot   = aux2 - NPOT*(*i);
	}
}
// -------------------------------------------------------------------------------
void Free_Arrays_n_And_n_v(parameters *par)
{
	free_ivector((*par).n_of_n_v, 0, (*par).n_2D-1);
	free_ivector((*par).n_v_of_n, 0, (*par).n_2D-1);
}
// -------------------------------------------------------------------------------
