#include "spheroidal.h"

//  -----------------------------------------------
double s_of_sig(parameters par, double sig)
{
	double del=par.del_s;

	if(fabs(del) < TINY) 
		return sig;
	else
		return 1.-asinh((1.-sig)*sinh(del))/del;
}
//  -----------------------------------------------
double t_of_tau(parameters par, double tau)
{
	double del=par.del_t;

	if(fabs(del) < TINY) 
		return tau;
	else
		return asinh(tau*sinh(del))/del;
}
//  -----------------------------------------------
double s_of_xi2(parameters par, double xi2)
{
	double rp = par.rp, re = par.re, xi = Sqrt(xi2), sig = re/(re-rp+xi);

	return s_of_sig(par, sig);
}
//   -----------------------------------------------
void get_sig_ArbitPos(parameters par, double s, double *sig, double *sig_m_1, double *sig_s)
{
	double del_s = par.del_s, aux = 1./sinch(del_s), sm1=s-1.;

		*sig_m_1 = sm1*sinch(del_s*sm1)*aux;
		*sig     = 1. + (*sig_m_1);
		*sig_s   = cosh(del_s*sm1)*aux;
}
//   -----------------------------------------------
void get_tau_ArbitPos(parameters par, double t, double *tau, double *tau_m_1, double *tau_t)
{
	double del_t = par.del_t, aux = 1./sinch(del_t);

		*tau     = t*    sinch(del_t*t)*aux;     // tau    = sinh(del*t)/sinh(del)
		*tau_m_1 = *tau - 1.;
		*tau_t   =       cosh( del_t*t)*aux;     // tau_t  = del*cosh(del*t)/sinh(del)
}
//  -----------------------------------------------
void coord_st_To_xy_ArbitPos(parameters par, int idom, double *cheb_G, double *cheb_dG, double s, double t, 
	coordinate *x, coordinate *y)
{// Calculation of the coordinates (x,y) from (s,t) along with first and second
 // derivatives with repect to s and t
	switch(idom){
		case 0: coord_0_ArbitPos(par, cheb_G, cheb_dG, s, t, x, y); break;
		case 1: coord_1_ArbitPos(par, cheb_G, cheb_dG, s, t, x, y); break;
	}
}
//   -----------------------------------------------
void coord_0_ArbitPos(parameters par, double *cheb_G, double *cheb_dG, double s, double t, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 0
	int  idom = 0, nt=par.nt[idom];
	double re   = par.re, xe = re*re, rp = par.rp, yp = rp*rp,
	       sig, sig_m_1, sig2, sig_s, sig2_s, tau, tau_m_1, tau_t, xi, xi_s, xi2, xi2_s;

	get_sig_ArbitPos(par, s, &sig, &sig_m_1, &sig_s);
	get_tau_ArbitPos(par, t, &tau, &tau_m_1, &tau_t);

	sig2    = sig*sig;
	sig2_s  = 2.*sig*sig_s;
	xi      = rp - re*sig_m_1/sig;
	xi_s    = -re*sig_s/sig2;
	xi2     = xi*xi;
	xi2_s   = 2.*xi*xi_s;

	(*x).d0  = (xi2 + xe - yp)*tau;
	(*x).d1  = xi2_s*tau;
	(*x).d2  = (xi2 + xe - yp)*tau_t;

	(*y).d0  =  fabs(tau_m_1*(yp-xi2) - xe*tau + chebev(0., 1., cheb_G, nt, t));
	(*y).d1  = -tau_m_1*xi2_s;
	(*y).d2  =  tau_t* (yp-xi2-xe) + chebev(0., 1., cheb_dG, nt, t);
}
//   -----------------------------------------------
void coord_1_ArbitPos(parameters par, double *cheb_G, double *cheb_dG, double s, double t, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 1
	int  idom = 0, nt=par.nt[idom];
	double re   = par.re, xe = re*re, tau, tau_m_1, tau_t;

	get_tau_ArbitPos(par, t, &tau, &tau_m_1, &tau_t);

	(*x).d0   = xe*tau;
	(*x).d2   = xe*tau_t;
	(*x).d1   = 0.;
	
	(*y).d1   = fabs(chebev(0., 1., cheb_G, nt, t)  - xe*tau);
	(*y).d2  =    s*(chebev(0., 1., cheb_dG, nt, t) - xe*tau_t);
	(*y).d0   = s*(*y).d1;
}
//  -----------------------------------------------
void xy_To_st(parameters par, double *cheb_G, double *cheb_dG,	double x, double y, int *idom, double *s, double *t)
{
	int nt=par.nt[NDOM-1], newt_iter, newt_max_iter = 10, idom0 = *idom;
	double re = par.re, xe = re*re, rp = par.rp, yp = rp*rp, s0 = *s, t0 = *t,
	       X, Y, XX, XX_s, XX_t, YY, YY_s, YY_t, arg, arg_s, arg_t,
	       f1, f2, f1_s, f1_t, f2_s, f2_t, det;
	coordinate xx, yy;

	if(y <= TINY){
		if(x > xe){
			*idom = 0;
			*s    = s_of_xi2(par, x + yp - xe);
			*t    = 1.;
		}
		else{
			*idom = 1;
			*s    = 0.;
			*t    = t_of_tau(par, x/xe);
		}
	}

	if(x <= TINY){
		*t    = 0.;
		if(y > yp) {
			*idom = 0;
			*s    = s_of_xi2(par, y);
		}
		else{
			*idom = 1;
			*s    = y/yp;
		}
	}

	if (x > TINY && y > TINY){
		*t = t_of_tau(par, x/xe);
		if( *t >= 0. && *t <= 1.){
			*s = y/(chebev(0., 1., cheb_G, nt, *t)-x);
			if( *s <= 1) *idom = 1;
			else         *idom = 0;
		}
		else
			*idom = 0;

		if(*idom == 0){
			*s = s0;
			*t = t0;
			X  = 1./(x/xe+y/yp);
			Y  = x*X/xe;

			if( *idom != idom0) *s = 1.;
			newt_iter = 0;
			do{
				newt_iter += 1;
				coord_st_To_xy_ArbitPos(par, *idom, cheb_G, cheb_dG, *s, *t, &xx, &yy);
				arg   = xx.d0/xe+yy.d0/yp;
				arg_s = xx.d1/xe+yy.d1/yp;
				arg_t = xx.d2/xe+yy.d2/yp;
				XX  = 1./arg;           XX_s = -XX*XX*arg_s;                   XX_t = -XX*XX*arg_t;
				YY  = xx.d0*XX/xe;      YY_s = (xx.d1*XX + xx.d0*XX_s)/xe;     YY_t = (xx.d2*XX + xx.d0*XX_t)/xe;
				f1  = XX - X;           f1_s = XX_s;                           f1_t = XX_t;
				f2  = YY - Y;           f2_s = YY_s;                           f2_t = YY_t;

				det = f1_s*f2_t - f1_t*f2_s;

/*				printf("newt_iter = %d \t x=%e y=%e \t fabs(f1)+fabs(f2) = %e \t s=%e t=%e\n", 
					newt_iter, x, y, fabs(f1)+fabs(f2), *s, *t);*/
		
				*s -= ( f2_t*f1 - f1_t*f2)/det;
				*t -= (-f2_s*f1 + f1_s*f2)/det;
			}
			while(fabs(f1)+fabs(f2) > 1.e-08 && newt_iter <= newt_max_iter);
/*			printf("\n");
			coord_st_To_xy_ArbitPos(par, *idom, cheb_G, cheb_dG, *s, *t, &xx, &yy);
			if(fabs(x) < 1.) f1  = xx.d0 - x;
			else             f1  = 1. - xx.d0/x;
			if(fabs(y) < 1.) f2  = yy.d0 - y;
			else             f2  = 1. - yy.d0/y;
			if(fabs(f1)+fabs(f2) > 1.e-08){
				printf("FAILURE IN 'xyz_To_st':\n x=%e y=%e TEST_x=%e TEST_y = %e s = %e t = %e Domain = %d\n",
					x, y, f1, f2, *s, *t, *idom);
				exit(1);
			}*/
// 			getchar();
		}
	}
	coord_st_To_xy_ArbitPos(par, *idom, cheb_G, cheb_dG, *s, *t, &xx, &yy);
	if(fabs(x) < 1.) f1  = xx.d0 - x;
	else             f1  = 1. - xx.d0/x;
	if(fabs(y) < 1.) f2  = yy.d0 - y;
	else             f2  = 1. - yy.d0/y;
	if(fabs(f1)+fabs(f2) > 1.e-08){
		printf("FAILURE IN 'xyz_To_st':\n x=%e y=%e x(s,t) - x = %e y(s,t) - y = %e s = %e t = %e Domain = %d\n",
			x, y, f1, f2, *s, *t, *idom);
		exit(1);
	}
}
// -------------------------------------------------------------------------------
double interpolate_w(parameters par, int ipot, double t, double *w)
{
	int nt=par.nt[NDOM-1], j;
	double p[NMAX];

	for(j=0;j<nt;j++) {
		p[j]=w[Index_Star(par,ipot,j)];
	}
	chebft_Extremes(p,nt,0);
	return chebev(0., 1., p, nt, t);
}
// -------------------------------------------------------------------------------
double interpolate_v(parameters par, int idom, int ipot, double s, double t, double *v)
{
	int nt=par.nt[idom], ns=par.ns[idom], i, j;
	double p[NMAX], q[NMAX];

	for(j=0;j<nt;j++) {
		for(i=0;i<ns;i++)
			p[i]=v[Index(par,idom,ipot,i,j)];
		chebft_Extremes(p,ns,0);
		q[j] = chebev(0., 1., p, ns, s);
	}
	chebft_Extremes(q,nt,0);
	return chebev(0., 1., q, nt, t);
}
// -------------------------------------------------------------------------------
void v_To_V_ArbitPos(parameters par, double *cheb_G, double *cheb_dG, 
	int idom, double s, double t, derivs_2D v, derivs_1D w, derivs_2D V)
{
	int ipot;
	coordinate xx, yy;

	for(ipot=0; ipot<NPOT; ipot++){
		if(idom == NDOM-1){
			V.d0[ipot]  = interpolate_w(par, ipot, t, w.d0) + interpolate_v(par, idom, ipot, s, t, v.d0);
			V.d1[ipot]  =                                     interpolate_v(par, idom, ipot, s, t, v.d1);
			V.d2[ipot]  = interpolate_w(par, ipot, t, w.d2) + interpolate_v(par, idom, ipot, s, t, v.d2);
		}
		else{
			V.d0[ipot]  = interpolate_v(par, idom, ipot, s, t, v.d0);
			V.d1[ipot]  = interpolate_v(par, idom, ipot, s, t, v.d1);
			V.d2[ipot]  = interpolate_v(par, idom, ipot, s, t, v.d2);
		}
	}
	coord_st_To_xy_ArbitPos(par, idom, cheb_G, cheb_dG, s, t, &xx, &yy);
	pot_st_To_xy(par, xx, yy, V);
}
//  -----------------------------------------------
void Get_cheb_G(parameters par, derivs_1D G, double *cheb_G, double *cheb_dG)
{
	int nt=par.nt[NDOM-1], j;

	for(j=0;j<nt;j++){
		cheb_G[j]  = G.d0[j];
		cheb_dG[j] = G.d2[j];
	}
	chebft_Extremes(cheb_G,  nt,0);
	chebft_Extremes(cheb_dG, nt,0);
}
//  -----------------------------------------------
void Interpolation_To_Cylindrical_Gridpoint(parameters par, double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V,
	double x, double y, int *idom, double *s, double *t, double *h, double *p, double *eps, double *muB)
{	
	xy_To_st(par, cheb_G, cheb_dG, x, y, idom, s, t);
	v_To_V_ArbitPos(par, cheb_G, cheb_dG, *idom, *s, *t, v, w, V);

	if(*idom == 1){
		double u  = V.d0[0],       B     = V.d0[1],   om = V.d0[2],
			  A1  = x*exp(-2.*u),  A2    = B*B*exp(2.*u),
			diff  = par.Omega - om,  diff2 = diff*diff, e2V = A2-A1*diff2;

		*h        = par.V0 - 0.5*log(e2V);     // h    = V0 - V
		eos_quantities(par, *h, p, eps, muB); // h(p) = Integrate[1/(eps(pp)+pp), {pp, 0, p}]
		*h        = par.h0*exp(*h);
	}
	else{
		*h = *p = *eps = *muB = 0.;
	}
}
//  -----------------------------------------------
void frintf_PhysQuant_CylCoord(parameters par, double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V)
{
	FILE *stream;
	double x, y, xmax = par.xmax, ymax = par.ymax, s, t, h, p, eps, muB;
	int i, j, nx = par.nx, ny = par.ny, idom;

	stream=fopen("PhysQuant_CylCoord.dat", "w");

	fprintf(stream,"#  (1):  x,  (2):    y,  \n");
	fprintf(stream,"#  (3):  u,  (4):  u_x,  (5): u_y, \n");
	fprintf(stream,"#  (6):  B,  (7):  B_x,  (8): B_y, \n");
	fprintf(stream,"#  (9): om, (10): om_x, (11): om_y,\n");
	fprintf(stream,"# (12): alpha, (13): alpha_x, (14): alpha_y \n");
	fprintf(stream,"# (15):  h, (16):    p, (17): eps,  (18): muB \n\n\n");
	
	printf("Calculation of physical quantities in the cylindrical Box. Please wait ... \n");
	for(i=0; i<=nx; i++){
		x = xmax*sqr(1.*i/nx);
		for(j=0; j<=ny; j++){
			y = ymax*sqr(1.*j/ny);
			Interpolation_To_Cylindrical_Gridpoint(par, cheb_G, cheb_dG, v, w, V, x, y, &idom, &s, &t, &h, &p, &eps, &muB);
			fprintf(stream,
				"%20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e\n",
				x,       y,
				V.d0[0], V.d1[0], V.d2[0],
				V.d0[1], V.d1[1], V.d2[1],
				V.d0[2], V.d1[2], V.d2[2],
				V.d0[3], V.d1[3], V.d2[3],
				h, p, eps, muB
			);
		}
		printf("%6.1lf%% done.\r", 100.*i/nx); 
		fflush(0);
		fprintf(stream,"\n");
	}
	printf("\n");
	printf("Please perform `gnuplot PhysQuant_CylCoord.plt' \n\n");
	printf("\n");
}
//  -----------------------------------------------
	
