#include "spheroidal.h"

//  -----------------------------------------------
void h_max(parameters *par, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V)
{
// Calculates Mass and Angular Momentum Integrals in domain (Ndom-1) 
// Integration in upper half-space, over (s,t) \in [0,1]^2
	int idom=NDOM-1, nt=(*par).nt[idom], i=0, j;
	double  Laplace[NPOT],  grad_grad[NPOT*NPOT],  x,  u,  B,  om,  A0, A1,  A2,  diff,  diff2,  e2V, 
		h_plane[NMAX], dh_plane[NMAX];
	coordinate xx, yy;

	for(j=0; j<nt; j++) {
		coord_st_To_xy(*par, idom, i, j, G, &xx, &yy);

		NonLinPotentials(*par, idom, i, j, xx, yy, v, w, G, V, Laplace, grad_grad);
		x  = xx.d0;
		u  = V.d0[0];      B  = V.d0[1];     om = V.d0[2];

		A1    = x*exp(-2.*u); 
		A2    = B*B*exp(2.*u);
		diff  = (*par).Omega - om;
		diff2 = diff*diff;
		e2V   = A2-A1*diff2;

		h_plane[j] = (*par).V0 - 0.5*log(e2V);  // h    = V0 - V
	}

	chebft_Extremes(h_plane,nt,0);
	chder(0., 1., h_plane,  dh_plane, nt);
	A0 = chebev(0., 1., dh_plane, nt, 0.);
	A1 = chebev(0., 1., dh_plane, nt, 1.);
	if (A0*A1 > 0.){
		(*par).h_max = chebev(0., 1., h_plane, nt, 0.); 
		(*par).t_h_max = 0.;
	}
	else{
		double t0=0., t_h_max = 0.5, t1=1., A;
		for(i=0; i<60; i++){
			A = chebev(0., 1., dh_plane, nt, t_h_max);		
			if (A*A0 > 0.) t0 = t_h_max;
			else           t1 = t_h_max;
			t_h_max = 0.5*(t0+t1);
		}
		(*par).h_max   = chebev(0., 1., h_plane, nt, t_h_max);
		(*par).t_h_max = t_h_max;
	}
	(*par).h_max = (*par).h0*exp((*par).h_max);

}
//  -----------------------------------------------
void D_hmax(parameters *par, parameters *Dpar,
       derivs_2D v,  derivs_1D w,  derivs_1D G,  derivs_2D V,
       derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV)
{
	int idom=NDOM-1, nt=(*par).nt[idom], i=0, j;
	double  Laplace[NPOT],  grad_grad[NPOT*NPOT],  x,  u,  B,  om,  A1,  A2,  diff,  diff2,  e2V;
	double DLaplace[NPOT], Dgrad_grad[NPOT*NPOT], Dx, Du, DB, Dom, DA1, DA2, Ddiff, Ddiff2, De2V, Dh_plane[NMAX];
	coordinate xx, yy, Dxx, Dyy;

	for(j=0; j<nt; j++) {
		Dcoord_st_To_xy(*par, *Dpar, idom, i, j, G, DG, &xx, &yy, &Dxx, &Dyy);
		
		LinPotentials(*par, idom, i, j, xx,  yy,  v,  w,  G,  V,  Laplace, grad_grad,
										Dxx, Dyy, Dv, Dw, DG, DV, DLaplace,  Dgrad_grad);

		x   = xx.d0;   Dx  = Dxx.d0;
		u   = V.d0[0]; B   = V.d0[1]; om = V.d0[2]; Du = DV.d0[0]; DB = DV.d0[1]; Dom = DV.d0[2];

		A1    = x*exp(-2.*u);                DA1    = exp(-2.*u)*(Dx - 2.*x*Du);
		A2    = B*B*exp(2.*u);               DA2    = 2.*B*exp(2.*u)*(DB+B*Du);
		diff  = (*par).Omega - om;           Ddiff  = (*Dpar).Omega - Dom;
		diff2 = diff*diff;                   Ddiff2 = 2.*diff*Ddiff;
		e2V   = A2-A1*diff2;                 De2V   = DA2-(DA1*diff2 + A1*Ddiff2);

		Dh_plane[j]   = (*Dpar).V0 - 0.5*De2V/e2V;
	}
	chebft_Extremes(Dh_plane,nt,0);
	(*Dpar).h_max   = chebev(0., 1., Dh_plane, nt, (*par).t_h_max);
	(*Dpar).h_max   = (*Dpar).h_max*(*par).h_max;
}
//  -----------------------------------------------
void M_and_J_in(parameters *par, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V)
{
// Calculates Mass and Angular Momentum Integrals in domain (Ndom-1) 
// Integration in upper half-space, over (s,t) \in [0,1]^2
	int idom=NDOM-1, ns=(*par).ns[idom], Ns=ns-1, nt=(*par).nt[idom], Nt=nt-1, isignum, i, j;
	double  M0s[NMAX],  Mps[NMAX],  Ms[NMAX],  Js[NMAX],  Ts[NMAX],
		   M0st[NMAX], Mpst[NMAX], Mst[NMAX], Jst[NMAX], Tst[NMAX],
		Laplace[NPOT], grad_grad[NPOT*NPOT], det;
	double hat_zeta,  x,  u,  B,  om,  mu,  A1,  A2,  diff,  diff2,  e2V,  e2mu,  hat_vel,  vel2,  h,  p,  eps,  muB,  aux;
	coordinate xx, yy;

	for(i=0; i<ns; i++){
		for(j=0; j<nt; j++) {
			double
				a = Pih*i/Ns, sa=sin(a), s=sa*sa,
				b = Pih*j/Nt, sb=sin(b), t=sb*sb;
			coord_st_To_xy(*par, idom, i, j, G, &xx, &yy);

			NonLinPotentials(*par, idom, i, j, xx, yy, v, w, G, V, Laplace, grad_grad);
			x  = xx.d0;
			u  = V.d0[0];  B   = V.d0[1];  om  = V.d0[2];  mu  = V.d0[3];

			A1    = x*exp(-2.*u); 
			A2    = B*B*exp(2.*u);
			diff  = (*par).Omega - om;
			diff2 = diff*diff;
			e2V   = A2-A1*diff2;
			e2mu  = exp(2.*mu);

			hat_vel  = exp(-2.*u)*diff/B;
			vel2     = x*hat_vel*hat_vel;
			h        = (*par).V0 - 0.5*log(e2V);
			eos_quantities(*par, h, &p, &eps, &muB); // h(p) = Integrate[1/(eps(pp)+pp), {pp, 0, p}]
			
			if (i > 0 && j < Nt)// Sqrt(x) = sqrt(abs(x))
				hat_zeta = Sqrt(yy.d0/(s*(1.-t))); // hat{zeta} = zeta/Sqrt[s*(1-t)]
			else{
				if(i == 0 && j  < Nt) hat_zeta = Sqrt(yy.d1/(1.-t));
				if(i  > 0 && j == Nt) hat_zeta = Sqrt(-yy.d2/s);
				if(i == 0 && j == Nt) hat_zeta = Sqrt(-yy.d12);
			}

			det     = xx.d2*yy.d1;
			aux     = e2mu*Pi/hat_zeta*det;
			M0st[j] = aux*muB*exp(-u)/Sqrt(1.-vel2);
			Mpst[j] = aux*eps *exp(-u)/Sqrt(1.-vel2);
			Jst[j]  = aux*exp(-2.*u)*x*hat_vel/(1.-vel2)*(eps+p);
			Mst[j]  = aux*B*((eps+p)*(1.+vel2)/(1.-vel2) + 2.*p) + 2.*om*Jst[j];
			Tst[j]  = 0.5*(*par).Omega*Jst[j];
		}
		chebft_Extremes(M0st,nt,0);
		chebft_Extremes(Mpst,nt,0);
		chebft_Extremes(Mst,nt,0);
		chebft_Extremes(Jst,nt,0);
		chebft_Extremes(Tst,nt,0);
		M0s[i] = Mps[i] = Ms[i] = Js[i] = Ts[i] = 0.;
		if (Nt % 2 == 0) isignum = 1;
		else             isignum = -1;
		for(j=Nt; j>0; j--){
			int denom = (2*j-1)*(2*j+1)*isignum;
			M0s[i] += M0st[j]/denom;
			Mps[i] += Mpst[j]/denom;
			Ms[i]  +=  Mst[j]/denom;
			Js[i]  +=  Jst[j]/denom;
			Ts[i]  +=  Tst[j]/denom;
			isignum = -isignum;
		}
		M0s[i] = M0st[0] - 2.*M0s[i];
		Mps[i] = Mpst[0] - 2.*Mps[i];
		Ms[i]  =  Mst[0] - 2.* Ms[i];
		Js[i]  =  Jst[0] - 2.* Js[i];
		Ts[i]  =  Tst[0] - 2.* Ts[i];
	}
	chebft_Extremes(M0s,ns,0);
	chebft_Extremes(Mps,ns,0);
	chebft_Extremes(Ms, ns,0);
	chebft_Extremes(Js, ns,0);
	chebft_Extremes(Ts, ns,0);
	(*par).M0 = (*par).Mp = (*par).M_in = (*par).J_in = (*par).T = 0.;
	for(i=Ns; i>0; i--){
		int denom = (2*i-1)*(2*i+1);
		(*par).M0   += M0s[i]/denom;
		(*par).Mp   += Mps[i]/denom;
		(*par).M_in +=  Ms[i]/denom;
		(*par).J_in +=  Js[i]/denom;
		(*par).T    +=  Ts[i]/denom;
	}
	(*par).M0   = M0s[0] - 2.*(*par).M0;
	(*par).Mp   = Mps[0] - 2.*(*par).Mp;
	(*par).M_in =  Ms[0] - 2.*(*par).M_in;
	(*par).J_in =  Js[0] - 2.*(*par).J_in;
	(*par).T    =  Ts[0] - 2.*(*par).T;
	(*par).W    =  (*par).Mp + (*par).T - (*par).M_in;
}
//  -----------------------------------------------
void D_M_and_J_in(parameters *par, parameters *Dpar,
       derivs_2D v,   derivs_1D w,  derivs_1D G, derivs_2D V,
       derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV)
{
	int idom=NDOM-1, ns=(*par).ns[idom], Ns=ns-1, nt=(*par).nt[idom], Nt=nt-1, isignum, i, j;
	double Laplace[NPOT], grad_grad[NPOT*NPOT], det,
		 hat_zeta,  x,  u,  B,  om,  mu,  A1,  A2,  diff,  diff2,  e2V,  e2mu,  hat_vel,  vel2,  h,  p,  eps,  muB,  aux;
	double DM0s[NMAX], DM0st[NMAX], DLaplace[NPOT], Dgrad_grad[NPOT*NPOT], Ddet,
		Dhat_zeta, Dx, Du, DB, Dom, Dmu, DA1, DA2, Ddiff, Ddiff2, De2V, De2mu, Dhat_vel, Dvel2, Dh, Dp, Deps, DmuB, Daux;
	coordinate xx, yy, Dxx, Dyy;

	for(i=0; i<ns; i++){
		for(j=0; j<nt; j++) {
			double
				a = Pih*i/Ns, sa=sin(a), s=sa*sa,
				b = Pih*j/Nt, sb=sin(b), t=sb*sb;
			Dcoord_st_To_xy(*par, *Dpar, idom, i, j, G, DG, &xx, &yy, &Dxx, &Dyy);
			
			LinPotentials(*par, idom, i, j, xx,  yy,  v,  w,  G,  V,  Laplace, grad_grad,
											Dxx, Dyy, Dv, Dw, DG, DV, DLaplace,  Dgrad_grad);

			x   = xx.d0;    Dx  = Dxx.d0;
			u   = V.d0[0];  B   = V.d0[1];  om  = V.d0[2];  mu  = V.d0[3];
			Du  = DV.d0[0]; DB  = DV.d0[1]; Dom = DV.d0[2]; Dmu = DV.d0[3];
	
			A1    = x*exp(-2.*u);                DA1       = exp(-2.*u)*(Dx - 2.*x*Du);
			A2    = B*B*exp(2.*u);               DA2       = 2.*B*exp(2.*u)*(DB+B*Du);
			diff  = (*par).Omega - om;             Ddiff     = (*Dpar).Omega - Dom;
			diff2 = diff*diff;                   Ddiff2    = 2.*diff*Ddiff;
			e2V   = A2-A1*diff2;                 De2V      = DA2-(DA1*diff2 + A1*Ddiff2);
			e2mu  = exp(2.*mu);                  De2mu     = 2.*e2mu*Dmu;

			hat_vel  = exp(-2.*u)*diff/B;        Dhat_vel  = exp(-2.*u)*(Ddiff - diff*(2.*Du + DB/B))/B;
			vel2     = x*hat_vel*hat_vel;        Dvel2     = hat_vel*(Dx*hat_vel + 2.*x*Dhat_vel);
			h        = (*par).V0 - 0.5*log(e2V); Dh        = (*Dpar).V0 - 0.5*De2V/e2V;
			eos_quantities(  *par, h, &p, &eps, &muB);
			D_eos_quantities(*par, h, Dh, &Dp, &Deps, &DmuB);

			if (i > 0 && j < Nt){
				hat_zeta  = Sqrt(yy.d0/(s*(1.-t))); // hat{zeta} = zeta/Sqrt[s*(1-t)]
				Dhat_zeta = hat_zeta*0.5*Dyy.d0/yy.d0;
			}
			else{
				if(i == 0 && j  < Nt){
					hat_zeta  = Sqrt(yy.d1/(1.-t));
					Dhat_zeta = hat_zeta*0.5*Dyy.d1/yy.d1;
				}
				if(i  > 0 && j == Nt){
					hat_zeta  = Sqrt(-yy.d2/s);
					Dhat_zeta = hat_zeta*0.5*Dyy.d2/yy.d2;
				}
				if(i == 0 && j == Nt){ 
					hat_zeta  = Sqrt(-yy.d12);
					Dhat_zeta = hat_zeta*0.5*Dyy.d12/yy.d12;
				}
			}
			det      = xx.d2*yy.d1;
			Ddet     = Dxx.d2*yy.d1+xx.d2*Dyy.d1;
			aux      = e2mu*Pi/hat_zeta*det;
			Daux     = aux*(De2mu/e2mu - Dhat_zeta/hat_zeta) + e2mu*Pi/hat_zeta*Ddet;
// 			M0st     = aux*muB*exp(-u)/Sqrt(1.-vel2);
			DM0st[j] = exp(-u)/Sqrt(1.-vel2)*(Daux*muB + aux*DmuB + aux*muB*(-Du + 0.5*Dvel2/(1.-vel2)));
		}
		chebft_Extremes(DM0st,nt,0);
		DM0s[i] = 0.;
		if (Nt % 2 == 0) isignum = 1;
		else             isignum = -1;
		for(j=Nt; j>0; j--){
			int denom = (2*j-1)*(2*j+1)*isignum;
			DM0s[i] += DM0st[j]/denom;
			isignum = -isignum;
		}
		DM0s[i] = DM0st[0] - 2.*DM0s[i];
	}
	chebft_Extremes(DM0s,ns,0);
	(*Dpar).M0 = 0.;
	for(i=Ns; i>0; i--){
		int denom = (2*i-1)*(2*i+1);
		(*Dpar).M0   += DM0s[i]/denom;
	}
	(*Dpar).M0   = DM0s[0] - 2.*(*Dpar).M0;
}
//  -----------------------------------------------
void M_and_J_out(parameters *par, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, int idom, int i)
{
// Calculates Mass and Angular Momentum Surface Integrals in domain (idom) (integration in upper half-space):

//  Integration over s=const=sin^2[Pih*i/Ns] : ['=d/dt]
//		M/2 = 1/2*Integrate[                         rho*B*[rho'(t)*nu_zeta    - zeta'(t)*nu_rho]
//		                    -omega/2*(rho*B)^3*exp[-4*nu]* [rho'(t)*omega_zeta - zeta'(t)*omega_rho],
//		                   {t,0,1}]
//		    = 1/2*Integrate[MInt[t]/Sqrt[1-t], {t,0,1}]
//		J/2 = -1/8*Integrate[(rho*B)^3*exp[-4*nu]*[rho'(t)*omega_zeta - zeta'(t)*omega_rho], {t,0,1}]
//		    = -1/8*Integrate[JInt[t]/Sqrt[1-t], {t,0,1}]
	int nt=(*par).nt[idom], Nt=nt-1, isignum, j;
	double M, J, MInt[NMAX], JInt[NMAX], Laplace[NPOT], grad_grad[NPOT*NPOT], x, y, hat_zeta,
		u, ux, uy, B, Bx, By, om, omx, omy, Bnux, Bnuy, aux_Bnu, aux_om;
	coordinate xx, yy;

	for(j=0; j<nt; j++){
		double b=Pih*j/Nt, sb=sin(b), t=sb*sb;
		coord_st_To_xy(*par, idom, i, j, G, &xx, &yy);
		NonLinPotentials(*par, idom, i, j, xx, yy, v, w, G, V, Laplace, grad_grad);
		x    = xx.d0;    y  = yy.d0;
		if(j < Nt) hat_zeta = Sqrt(y/(1.-t)); // hat{zeta} = zeta/Sqrt[1-t]
		else       hat_zeta = Sqrt(-yy.d2);
		
		u    = V.d0[0]; ux  = V.d1[0];  uy = V.d2[0];
		B    = V.d0[1];  Bx = V.d1[1];  By = V.d2[1];
		om   = V.d0[2]; omx = V.d1[2]; omy = V.d2[2];
		Bnux = B*ux + Bx;
		Bnuy = B*uy + By;

		aux_Bnu  = xx.d2*hat_zeta*(1.-t)*Bnuy - yy.d2/hat_zeta*x*Bnux;
		aux_om   = xx.d2*hat_zeta*(1.-t)*omy  - yy.d2/hat_zeta*x*omx;
		
		JInt[j]  = -0.25*x*exp(-4.*u)/B*aux_om;
		MInt[j]  = aux_Bnu  + 2.*om*JInt[j];
	}
	chebft_Extremes(MInt,nt,0);
	chebft_Extremes(JInt,nt,0);
	M = J = 0.;
	if (Nt % 2 == 0) isignum = 1;
	else             isignum = -1;
	for(j=Nt; j>0; j--){
		int denom = (2*j-1)*(2*j+1)*isignum;
		M += MInt[j]/denom;
		J += JInt[j]/denom;
		isignum = -isignum;
	}
	(*par).M = MInt[0] - 2.*M;
	(*par).J = JInt[0] - 2.*J;
}
//  -----------------------------------------------
void D_M_and_J_out(parameters *par, parameters *Dpar,
       derivs_2D v,   derivs_1D w,  derivs_1D G, derivs_2D V,
		derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, int idom, int i)
{
	int nt=(*par).nt[idom], Nt=nt-1, isignum, j;
	double DM, DJ, DMInt[NMAX], DJInt[NMAX],
		Laplace[NPOT], grad_grad[NPOT*NPOT], DLaplace[NPOT], Dgrad_grad[NPOT*NPOT], JInt,
		x,  y,  hat_zeta,  u,  ux,  uy,  B,  Bx,  By,  om,  omx,  omy,  Bnux,  Bnuy,  aux_Bnu,  aux_om,
		Dx, Dy, Dhat_zeta, Du, Dux, Duy, DB, DBx, DBy, Dom, Domx, Domy, DBnux, DBnuy, Daux_Bnu, Daux_om;
	coordinate xx, yy, Dxx, Dyy;

	for(j=0; j<nt; j++){
		double b=Pih*j/Nt, sb=sin(b), t=sb*sb;
		Dcoord_st_To_xy(*par, *Dpar, idom, i, j, G, DG, &xx, &yy, &Dxx, &Dyy);
		
		LinPotentials(*par, idom, i, j, xx,  yy,  v,  w,  G,  V,  Laplace, grad_grad,
										Dxx, Dyy, Dv, Dw, DG, DV, DLaplace,  Dgrad_grad);
		x    =  xx.d0;    y  =  yy.d0;
		Dx   = Dxx.d0;   Dy  = Dyy.d0;
		if(j < Nt) {// hat{zeta} = zeta/Sqrt[1-t]
			hat_zeta  = Sqrt(y/(1.-t)); 
			Dhat_zeta = hat_zeta*0.5*Dy/y;
		}
		else{
			hat_zeta  = Sqrt(-yy.d2);
			Dhat_zeta = hat_zeta*0.5*Dyy.d2/yy.d2;
		}
		
		u    = V.d0[0];  ux = V.d1[0];  uy = V.d2[0]; Du    = DV.d0[0];  Dux = DV.d1[0]; Duy  = DV.d2[0];
		B    = V.d0[1];  Bx = V.d1[1];  By = V.d2[1]; DB    = DV.d0[1];  DBx = DV.d1[1]; DBy  = DV.d2[1];
		om   = V.d0[2]; omx = V.d1[2]; omy = V.d2[2]; Dom   = DV.d0[2]; Domx = DV.d1[2]; Domy = DV.d2[2];
		Bnux = B*ux + Bx;                             DBnux = DB*ux + B*Dux + DBx;
		Bnuy = B*uy + By;                             DBnuy = DB*uy + B*Duy + DBy;

		aux_Bnu  = xx.d2*hat_zeta*(1.-t)*Bnuy - yy.d2/hat_zeta*x*Bnux;
		aux_om   = xx.d2*hat_zeta*(1.-t)*omy  - yy.d2/hat_zeta*x*omx;
		Daux_Bnu = (1.-t)*(Dxx.d2*hat_zeta*Bnuy + xx.d2*Dhat_zeta*Bnuy + xx.d2*hat_zeta*DBnuy)
			-(Dyy.d2*x*Bnux + yy.d2*Dx*Bnux + yy.d2*x*DBnux - yy.d2/hat_zeta*x*Bnux*Dhat_zeta)/hat_zeta;
		Daux_om  = (1.-t)*(Dxx.d2*hat_zeta*omy  + xx.d2*Dhat_zeta*omy  + xx.d2*hat_zeta*Domy)
			-(Dyy.d2*x*omx  + yy.d2*Dx*omx  + yy.d2*x*Domx  - yy.d2/hat_zeta*x*omx*Dhat_zeta )/hat_zeta;

		JInt     = -0.25*exp(-4.*u)/B*x*aux_om;
		DJInt[j] = -0.25*exp(-4.*u)/B*(Dx*aux_om + x*Daux_om - x*aux_om*(4.*Du+DB/B));
		
		// MInt  = aux_Bnu  + 2.*om*JInt;
		DMInt[j] = Daux_Bnu + 2.*(Dom*JInt + om*DJInt[j]);
	}
	chebft_Extremes(DMInt,nt,0);
	chebft_Extremes(DJInt,nt,0);
	DM = DJ = 0.;
	if (Nt % 2 == 0) isignum = 1;
	else             isignum = -1;
	for(j=Nt; j>0; j--){
		int denom = (2*j-1)*(2*j+1)*isignum;
		DM += DMInt[j]/denom;
		DJ += DJInt[j]/denom;
		isignum = -isignum;
	}
	(*Dpar).M = DMInt[0] - 2.*DM;
	(*Dpar).J = DJInt[0] - 2.*DJ;
}	
//  -----------------------------------------------
double ShedParameter(parameters par, derivs_1D G)
{
	int Ns=par.ns[NDOM-1]-1, Nt=par.nt[NDOM-1]-1, i=Ns, j=Nt;
	double 	yp = G.d0[0], xe = G.d0[Nt], beta;
	coordinate xx, yy;

	coord_st_To_xy(par, NDOM-1, i, j, G, &xx, &yy);

	beta  = -xe/yp*yy.d2/xx.d2;
	
	return beta;
}
//  -----------------------------------------------
double D_ShedParameter(parameters par, parameters Dpar, derivs_1D G, derivs_1D DG)
{
	int Ns=par.ns[NDOM-1]-1, Nt=par.nt[NDOM-1]-1, i=Ns, j=Nt;
	double 	yp = G.d0[0], xe = G.d0[Nt], Dyp = DG.d0[0], Dxe = DG.d0[Nt], beta, Dbeta;
	coordinate xx, yy, Dxx, Dyy;

	Dcoord_st_To_xy(par, Dpar, NDOM-1, i, j, G, DG, &xx, &yy, &Dxx, &Dyy);

	beta  = -xe/yp*yy.d2/xx.d2;
	Dbeta = beta*(Dxe/xe - Dyp/yp - Dxx.d2/xx.d2) - xe/yp*Dyy.d2/xx.d2;
	
	return Dbeta;
}
//  -----------------------------------------------
void CenterParameters(parameters *par, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V)
{
	int idom=NDOM-1, i=0, j=0;
	double Laplace[NPOT], grad_grad[NPOT*NPOT], u, B, e2V, h;
	coordinate xx, yy;

	coord_st_To_xy(*par, idom, i, j, G, &xx, &yy);
	NonLinPotentials(*par, idom, i, j, xx, yy, v, w, G, V, Laplace, grad_grad);
	u              = V.d0[0];
	B              = V.d0[1];
	e2V            = B*B*exp(2.*u);
	h              = (*par).V0 - 0.5*log(e2V);  //  h    = V0 - V - B_Om(Y), B_Om=0 for Y=0, i.e. rho=0
	(*par).u_c     = u;
	(*par).B_c     = B;
	(*par).om_c    = V.d0[2];
	(*par).mu_c    = V.d0[3];
	eos_quantities(*par, h, &((*par).p_c), &((*par).eps_c), &((*par).muB_c));
	(*par).h_c     = (*par).h0*exp(h);
}
//  -----------------------------------------------
void D_CenterParameters(parameters *par, parameters *Dpar,
       derivs_2D v,   derivs_1D w,  derivs_1D G,  derivs_2D V,
       derivs_2D Dv, derivs_1D Dw, derivs_1D DG,  derivs_2D DV)
{
	int idom=NDOM-1, i=0, j=0;
	double Laplace[NPOT], grad_grad[NPOT*NPOT], DLaplace[NPOT], Dgrad_grad[NPOT*NPOT],
			u,  B,  e2V, h, Du, DB, De2V, Dh;
	coordinate xx, yy, Dxx, Dyy;

	Dcoord_st_To_xy(*par, *Dpar, idom, i, j, G, DG, &xx, &yy, &Dxx, &Dyy);
	LinPotentials(*par, idom, i, j, xx,  yy,  v,  w,  G,  V,  Laplace,   grad_grad,
									Dxx, Dyy, Dv, Dw, DG, DV, DLaplace,  Dgrad_grad);
	u              = V.d0[0];           Du   = DV.d0[0];
	B              = V.d0[1];           DB   = DV.d0[1];
	e2V            = B*B*exp(2.*u);     De2V = 2.*e2V*(DB/B+Du);
	h              = (*par).V0  - 0.5*log(e2V);
	Dh             = (*Dpar).V0 - 0.5*De2V/e2V;
	D_eos_quantities(*par, h, Dh, &((*Dpar).p_c), &((*Dpar).eps_c), &((*Dpar).muB_c));
	(*Dpar).h_c    = (*par).h_c*Dh;
	(*Dpar).B_c    = DB;
}
//  -----------------------------------------------
void Rcirc(parameters *par, derivs_2D v, derivs_1D w, derivs_2D V)
{
	int idom=NDOM-1, i=0, j=(*par).nt[idom]-1;

	v_To_V(*par, idom, i, j, v, w, V);

	(*par).Rcirc = (*par).re*exp(-V.d0[0]);
}
//  -----------------------------------------------
void D_Rcirc(parameters *par, parameters *Dpar, 
	derivs_2D v, derivs_1D w, derivs_2D V, derivs_2D Dv, derivs_1D Dw, derivs_2D DV)
{
	int idom=NDOM-1, i=0, j=(*par).nt[idom]-1;

	v_To_V(*par, idom, i, j,  v,  w,  V);
	v_To_V(*par, idom, i, j, Dv, Dw, DV);

	(*Dpar).Rcirc = exp(-V.d0[0])*((*Dpar).re - (*par).re*DV.d0[0]);
}
//  -----------------------------------------------
void GetPhysicalParameters_NonLin(parameters *parX, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V)
{	
	int idom, i, Nt=(*parX).nt[NDOM-1]-1;
	double sig, sig_m_1, sig_s, sig_ss, M_Infinity,
		yp = G.d0[0],  xe = G.d0[Nt],
		rp = sqrt(yp), re = sqrt(xe);
		
	idom=0; // surface integrals for M and J taken in domain 0
	M_and_J_out(parX, v, w, G, V, idom, (*parX).ns[idom]-1);
	
	M_and_J_in( parX, v, w, G, V);

	(*parX).rp      = rp;
	(*parX).re      = re;
	(*parX).r_ratio = rp/re;
	(*parX).beta    = ShedParameter(*parX, G);

	CenterParameters(parX, v, w, G, V);
	h_max(parX, v, w, G, V);

	Rcirc(parX, v, w, V);

	idom=0;   i=0;
	get_sig(*parX, idom, i, &sig, &sig_m_1, &sig_s, &sig_ss);
	M_Infinity      = -re*v.d1[Index(*parX, 0, 0, 0, 0)]/sig_s;
	(*parX).M_test  = 1.-(*parX).M/M_Infinity;

	(*parX).Eb         = (*parX).M0 - (*parX).M;
}
//  -----------------------------------------------
void GetPhysicalParameters_Lin(parameters *parX, parameters *DparX,
       derivs_2D v,   derivs_1D w,  derivs_1D G,  derivs_2D V,
       derivs_2D Dv, derivs_1D Dw, derivs_1D DG,  derivs_2D DV)
{	
	int idom, Nt=(*parX).nt[NDOM-1]-1, ipar;
	double
		yp = G.d0[0],  xe = G.d0[Nt], Dyp = DG.d0[0],   Dxe = DG.d0[Nt],
		rp = sqrt(yp), re = sqrt(xe), Drp = 0.5*Dyp/rp, Dre = 0.5*Dxe/re;
	
	for(ipar=0; ipar<NPAR; ipar++){
		int choice = (*parX).choice[ipar];
		if (choice == CHOICE_M || choice == CHOICE_J|| choice == CHOICE_Eb){
			idom=0;  // surface integrals for M and J taken in domain 0
			D_M_and_J_out(parX, DparX, v, w, G, V, Dv, Dw, DG, DV, idom, (*parX).ns[idom]-1);
		}
		if (choice == CHOICE_M0 || choice == CHOICE_Eb){
			D_M_and_J_in(parX, DparX, v, w, G, V, Dv, Dw, DG, DV);
			if (choice == CHOICE_Eb) (*DparX).Eb = (*DparX).M0 - (*DparX).M;
		}
		if (choice == CHOICE_h_max)
			D_hmax(parX, DparX, v, w, G, V, Dv, Dw, DG, DV);
	}

	(*DparX).rp         = Drp;
	(*DparX).re         = Dre;
	(*DparX).r_ratio    = (Drp - Dre*rp/re)/re;
	(*DparX).beta       = D_ShedParameter(*parX, *DparX, G, DG);

	D_CenterParameters(parX, DparX, v, w, G, V, Dv, Dw, DG, DV);
	D_Rcirc(parX, DparX, v, w, V, Dv, Dw, DV);
}
//  -----------------------------------------------
void NonLinParamEqns(parameters par, parameters parX, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F)
{
	int n_2D=par.n_2D, nt=par.nt[NDOM-1], ipar;

	GetPhysicalParameters_NonLin(&parX, v, w, G, V);
	
	for(ipar=0; ipar<NPAR; ipar++){
		switch(par.choice[ipar]){
			case CHOICE_M         : F[n_2D + nt + ipar] = parX.M       - par.M;       break;
			case CHOICE_M0        : F[n_2D + nt + ipar] = parX.M0      - par.M0;      break;
			case CHOICE_J         : F[n_2D + nt + ipar] = parX.J       - par.J;       break;
			case CHOICE_rp        : F[n_2D + nt + ipar] = parX.rp      - par.rp;      break;
			case CHOICE_re        : F[n_2D + nt + ipar] = parX.re      - par.re;      break;
			case CHOICE_r_ratio   : F[n_2D + nt + ipar] = parX.r_ratio - par.r_ratio; break;
			case CHOICE_beta      : F[n_2D + nt + ipar] = parX.beta    - par.beta;    break;
			case CHOICE_V0        : F[n_2D + nt + ipar] = parX.V0      - par.V0;      break;
			case CHOICE_Omega     : F[n_2D + nt + ipar] = parX.Omega   - par.Omega;   break;
			case CHOICE_p_c       : F[n_2D + nt + ipar] = parX.p_c     - par.p_c;     break;
			case CHOICE_eps_c     : F[n_2D + nt + ipar] = parX.eps_c   - par.eps_c;   break;
			case CHOICE_muB_c     : F[n_2D + nt + ipar] = parX.muB_c   - par.muB_c;   break;
			case CHOICE_h_max     : F[n_2D + nt + ipar] = parX.h_max   - par.h_max;   break;
			case CHOICE_B_c       : F[n_2D + nt + ipar] = parX.B_c     - par.B_c;     break;
			case CHOICE_Eb        : F[n_2D + nt + ipar] = parX.Eb      - par.Eb;      break;
			case CHOICE_Rcirc     : F[n_2D + nt + ipar] = parX.Rcirc   - par.Rcirc;   break;
		}
	}
}
//  -----------------------------------------------
void LinParamEqns(parameters par, parameters parX, parameters DparX, 
					derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V,
					derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX)
{
	int n_2D=par.n_2D, nt=par.nt[NDOM-1], ipar;

	GetPhysicalParameters_Lin(&parX, &DparX, v, w, G, V, Dv, Dw, DG, DV);

	for(ipar=0; ipar<NPAR; ipar++){
		switch(par.choice[ipar]){
			case CHOICE_M         : JDX[n_2D + nt + ipar] = DparX.M;        break;
			case CHOICE_M0        : JDX[n_2D + nt + ipar] = DparX.M0;       break;
			case CHOICE_J         : JDX[n_2D + nt + ipar] = DparX.J;        break;
			case CHOICE_rp        : JDX[n_2D + nt + ipar] = DparX.rp;       break;
			case CHOICE_re        : JDX[n_2D + nt + ipar] = DparX.re;       break;
			case CHOICE_r_ratio   : JDX[n_2D + nt + ipar] = DparX.r_ratio;  break;
			case CHOICE_beta      : JDX[n_2D + nt + ipar] = DparX.beta;     break;
			case CHOICE_V0        : JDX[n_2D + nt + ipar] = DparX.V0;       break;
			case CHOICE_Omega     : JDX[n_2D + nt + ipar] = DparX.Omega;    break;
			case CHOICE_p_c       : JDX[n_2D + nt + ipar] = DparX.p_c;      break;
			case CHOICE_eps_c     : JDX[n_2D + nt + ipar] = DparX.eps_c;    break;
			case CHOICE_muB_c     : JDX[n_2D + nt + ipar] = DparX.muB_c;    break;
			case CHOICE_h_max     : JDX[n_2D + nt + ipar] = DparX.h_max;    break;
			case CHOICE_B_c       : JDX[n_2D + nt + ipar] = DparX.B_c;      break;
			case CHOICE_Eb        : JDX[n_2D + nt + ipar] = DparX.Eb;       break;
			case CHOICE_Rcirc     : JDX[n_2D + nt + ipar] = DparX.Rcirc;    break;
		}
	}
}
//  -----------------------------------------------
void Print_Parameters(parameters par)
{
	double p_max, eps_max, muB_max;

	printf(" Parameters:\n");
	printf(" -----------\n\n");
	
	printf(" M       = %20.15e \t J       = %20.15e \n",   par.M,           par.J);
//	printf(" M_in    = %20.15e \t J_in    = %20.15e \n",   par.M_in,        par.J_in);
//	printf(" Mp      = %20.15e \t T       = %20.15e \n",   par.Mp,          par.T);
//	printf(" M0      = %20.15e \t W       = %20.15e \n",   par.M0,          par.W);
//	printf(" T/W     = %20.15e \t Rcirc   = %20.15e \n",   par.T/par.W,  par.Rcirc);
	printf(" M0      = %20.15e \t Rcirc   = %20.15e \n",   par.M0,          par.Rcirc);
	printf(" Ebind   = %20.15e \t Omega   = %20.15e \n\n", par.Eb,          par.Omega);

	printf(" z       = %20.15e \t V0      = %20.15e \n",   exp(-par.V0)-1., par.V0);
	printf(" rp      = %20.15e \t re      = %20.15e \n",   par.rp,      par.re);
	printf(" r_ratio = %20.15e \t beta    = %20.15e \n",   par.r_ratio, par.beta );

	printf(" eps_c   = %20.15e \t p_c     = %20.15e \n",   par.eps_c,   par.p_c);
	printf(" muB_c   = %20.15e \t h_c     = %20.15e \n\n", par.muB_c,  par.h_c);

/*	printf(" t_h_max = %20.15e \t p_beta  = %20.15e \n",
		par.t_h_max, par.beta/(1.+par.beta));*/
	eos_quantities(par, log(par.h_max/par.h0), &p_max, &eps_max, &muB_max);
	printf(" eps_max = %20.15e \t p_max   = %20.15e \n",   eps_max, p_max);
	printf(" muB_max = %20.15e \t h_max   = %20.15e \n",   muB_max, par.h_max);
	printf(" u_c     = %20.15e \t B_c     = %20.15e \n",   par.u_c,     par.B_c);
	printf(" alpha_c = %20.15e \t om_c    = %20.15e \n\n", par.mu_c,    par.om_c);

	printf(" Resolution:  ns = nt = %d in domains 0 and 1\n\n",   par.ns[0]);
	
	display_eos_parameters(par);

	printf("\n Tests: (i) |1-M_infinity/M| = %e \t (ii) |alpha_c + mu_c|= %e\n", 
		fabs(par.M_test), fabs(par.u_c+par.mu_c));

	printf("\n");
	printf(" del_s = %e \n", par.del_s);
	printf("\n");
}
//  -----------------------------------------------
