#include "spheroidal.h"

// Scanning and printing data files.
//   -----------------------------------------------
void ScanInitialData(char *name, parameters *parInit, double **X)
{  /* A file called "name" is read.
     Information is stored in parInit (essentially bookkeeping)
	  Physical information is stored in X 
	*/ 
	FILE *stream_init;
	char string[200];
	int n, n_v, idom, ipot, ipar, ns, i, nt, j;
	double d_aux;	// i_aux, d_aux: int/double auxiliary variables

	stream_init = fopen(name, "r");
	
	(*parInit).nst[0] = 0;
	fscanf(stream_init, " %d %s ",  &ns, string);
	nt = ns;
	for(idom=0; idom<NDOM; idom++){
		(*parInit).ns[idom] = ns;
		(*parInit).nt[idom] = nt;
		(*parInit).nst[idom+1] = (*parInit).nst[idom] + ns*nt;
	}
	Get_Arrays_n_And_n_v(parInit);

	fscanf(stream_init, " %lf %s ",  &d_aux, string);    (*parInit).del_s  = d_aux;
// 	fscanf(stream_init, " %lf %s ",  &d_aux, string);    (*parInit).del_t  = d_aux;
	(*parInit).del_t  = 0.;

	scan_eos_parameters(   parInit, stream_init);

	*X  = dvector(0, (*parInit).ntotal-1);

	for(idom=0; idom<NDOM; idom++){ // Reading the Potential values
		ns=(*parInit).ns[idom];
		nt=(*parInit).nt[idom];
		for(ipot=0; ipot<NPOT; ipot++){
			for(j=0;j<nt;j++) {
				for(i=0;i<ns;i++){
					n_v = Index(*parInit, idom,ipot,i,j);
					n   = (*parInit).n_of_n_v[n_v];
					if(n>=0){
						fscanf(stream_init, " %lf %s ", &d_aux, string);
						(*X)[n] = d_aux;
					}
				}
			}
		}
	}
	nt=(*parInit).nt[NDOM-1];
	for(j=0;j<nt;j++) {   // Reading the Surface values
		fscanf(stream_init, " %lf %s ", &d_aux, string);
		(*X)[(*parInit).n_2D + j]   = d_aux;
	}
	
	for(ipar=0; ipar<NPAR; ipar++){     // Reading the Parameter values
		fscanf(stream_init, " %lf %s ", &d_aux, string);
		(*X)[(*parInit).n_2D + nt + ipar] = d_aux;
	}
	fclose(stream_init);
}
//   -----------------------------------------------
void ScanConfig(parameters *parInit, parameters *parGoal)
{
	FILE *stream_config;
	char string[200];
	int i_aux, idom, ipar, ns, nt, nx, ny;  // i_aux: an int auxiliary variable
	double d_aux, rhomax, zetamax;               // d_aux: a double auxiliary variable

	stream_config = fopen("Config", "r");

	fscanf(stream_config, " %s  %s",  (*parInit).SeqName, string);
	
// 	printf("  SeqName=%s \t   string = %s \n",    (*parInit).SeqName,     string);
// 	exit(1);

// 	fscanf(stream_config, " %d  %s ",  &i_aux, string);		(*parInit).Newton_itmin     = i_aux;
	(*parInit).Newton_itmin = 0;
	fscanf(stream_config, " %d  %s ",  &i_aux, string);		(*parInit).Newton_itmax     = i_aux;
	fscanf(stream_config, " %lf %s ",  &d_aux, string);		(*parInit).Newton_tol       = d_aux;
// 	fscanf(stream_config, " %d  %s ",  &i_aux, string);		(*parInit).Newton_verb      = i_aux;
	(*parInit).Newton_verb  = 1;

// 	fscanf(stream_config, " %d  %s ",  &i_aux, string);		(*parInit).bicgstab_itmax   = i_aux;
// 	fscanf(stream_config, " %lf %s ",  &d_aux, string);		(*parInit).bicgstab_decr    = d_aux;
// 	fscanf(stream_config, " %d  %s ",  &i_aux, string);		(*parInit).bicgstab_verb    = i_aux;
// 	
// 	fscanf(stream_config, " %lf %s ",  &d_aux, string);		(*parInit).PreCond_Improve  = d_aux;
// 	fscanf(stream_config, " %d  %s ",  &i_aux, string);		(*parInit).PreCond_itmax    = i_aux;
// 	fscanf(stream_config, " %d  %s ",  &i_aux, string);		(*parInit).PreCond_FD_Order = i_aux;

	(*parInit).bicgstab_itmax   = 40;
	(*parInit).bicgstab_decr    = 1.e-05;
	(*parInit).bicgstab_verb    = 0;
	
	(*parInit).PreCond_Improve  = 1.e-05;
	(*parInit).PreCond_itmax    = 6;
	(*parInit).PreCond_FD_Order = 4;

	*parGoal = *parInit;
	(*parGoal).nst[0] = 0;
	fscanf(stream_config, " %d %s ",  &ns, string);
	nt = ns;
	for(idom=0; idom<NDOM; idom++){
		(*parGoal).ns[idom] = ns;
		(*parGoal).nt[idom] = nt;
		(*parGoal).nst[idom+1] = (*parGoal).nst[idom] + ns*nt;
	}
	Get_Arrays_n_And_n_v(parGoal);

	fscanf(stream_config, " %s ",  string);
	for(ipar=0; ipar<NPAR; ipar++){
		fscanf(stream_config, " %d  %s ",  &i_aux, string); 
		(*parInit).choice[ipar] = (*parGoal).choice[ipar] = i_aux;
	}

	for(ipar=0; ipar<NPAR; ipar++){
		fscanf(stream_config, " %lf %s ",  &d_aux, string);
		switch((*parGoal).choice[ipar]){
			case CHOICE_M       :(*parGoal).M       = d_aux;     break;
			case CHOICE_M0      :(*parGoal).M0      = d_aux;     break;
			case CHOICE_J       :(*parGoal).J       = d_aux;     break;

			case CHOICE_rp      :(*parGoal).rp      = d_aux;     break;
			case CHOICE_re      :(*parGoal).re      = d_aux;     break;
			case CHOICE_r_ratio :(*parGoal).r_ratio = d_aux;     break;
			case CHOICE_beta    :(*parGoal).beta    = d_aux;     break;

			case CHOICE_V0      :(*parGoal).V0      = d_aux;     break;
			case CHOICE_Omega   :(*parGoal).Omega   = d_aux;     break;
			case CHOICE_p_c     :(*parGoal).p_c     = d_aux;     break;
			case CHOICE_eps_c   :(*parGoal).eps_c   = d_aux;     break;
			case CHOICE_muB_c   :(*parGoal).muB_c   = d_aux;     break;
			case CHOICE_h_max   :(*parGoal).h_max   = d_aux;     break;
			case CHOICE_B_c     :(*parGoal).B_c     = d_aux;     break;
			case CHOICE_Eb      :(*parGoal).Eb      = d_aux;     break;
			case CHOICE_Rcirc   :(*parGoal).Rcirc   = d_aux;     break;
		}
	}
	fscanf(stream_config, " %lf %s ",  &d_aux, string);  (*parGoal).del_s    = d_aux;
// 	fscanf(stream_config, " %lf %s ",  &d_aux, string);  (*parGoal).del_t    = d_aux;
	(*parGoal).del_t    = 0.;

	fscanf(stream_config, " %d  %s ",  &i_aux, string);
	(*parInit).n_Seq = (*parGoal).n_Seq = i_aux;

	fscanf(stream_config, " %d %d %s ",  &nx, &ny, string);
	(*parInit).nx   = (*parGoal).nx   = nx;
	(*parInit).ny   = (*parGoal).ny   = ny;
	fscanf(stream_config, " %lf %lf %s ",  &rhomax, &zetamax, string);  
	(*parInit).xmax = (*parGoal).xmax = rhomax*rhomax;
	(*parInit).ymax = (*parGoal).ymax = zetamax*zetamax;

	fscanf(stream_config, " %d  %s ",  &i_aux, string);

	if(i_aux != (*parInit).eos){
		printf(" Equation of state specified in `Config' does not coincide with that of the initial configuration.");
		printf(" Now exiting to system.\n\n");
		exit(1);
	}
	else{ 
		switch((*parInit).eos){
			case 0: // for homogeneous body
				{
//					fscanf(stream_config, " %lf %s ",  &d_aux, string);    (*parGoal).eps0  = d_aux;
					(*parInit).eps0 = (*parGoal).eps0  = 1.;
					(*parGoal).h0  = 1.;
					break;
				}
			case 1: // for polytropic body
				{
					fscanf(stream_config, " %lf %s ",  &d_aux, string);    (*parGoal).N_poly  = d_aux;
//					fscanf(stream_config, " %lf %s ",  &d_aux, string);    (*parGoal).K_poly  = d_aux;
					(*parInit).K_poly = (*parGoal).K_poly  = 1.;
					(*parGoal).h0  = 1.;
					break;
				}
			case 2: // for ideal, relativistic neutron gas,
					// normalization: for p->0 one finds the polytrope with n=3/2
				{
					(*parGoal).h0  = 1.;
					break;
				}
			case 3:// customized equation of state
				{
					int k, Np_Init = (*parInit).eos3_Np, Np_Goal;
			
					fill0_dvector((*parGoal).eos3_p, 0, NpMAX-1);
			
//					fscanf(stream_config, " %s ",  string);
					fscanf(stream_config, " %d  %lf %s ",  &Np_Goal, &d_aux, string); 
					(*parGoal).h0  = d_aux;
			
					for(k=1; k <= Np_Goal; k++){
						fscanf(stream_config, " %lf %s ",  &d_aux, string);    (*parGoal).eos3_p[k] = d_aux;
			// 					printf("k=%d \t p[k]=%e\n", k ,(*par).eos3_p[k]);
					}
					(*parGoal).eos3_Np = (*parInit).eos3_Np = maximum2(Np_Init, Np_Goal);
				}
		}
	}
	fclose(stream_config);
}
//   -----------------------------------------------
void ScanFiles(char *name, parameters *parInit, double **X, parameters *parGoal)
{
	ScanInitialData(name, parInit, X);
	ScanConfig(parInit, parGoal);
}
//   -----------------------------------------------
void PrintToFile(parameters par, double *X, int i_Seq)
{
	char name[50];
	FILE *stream;
	int n_2D=par.n_2D, idom, ipot, i, j, ns, Ns, nt, Nt, n, n_v;

	if (i_Seq <  10)              sprintf(name, "%s.00%1d", par.SeqName, i_Seq);
	if (i_Seq >= 10 && i_Seq<100) sprintf(name, "%s.0%2d",  par.SeqName, i_Seq);
	if (i_Seq >= 100)             sprintf(name, "%s.%3d",   par.SeqName, i_Seq);

	stream=fopen(name, "w");
	
	fprintf(stream, "%d \t ns_and_nt\n", par.ns[0]);
/*	for(idom=0; idom<NDOM; idom++)
		fprintf(stream, "%d \t %d \t ns_and_nt_domain_%1d\n", par.ns[idom], par.nt[idom],idom);*/
	
	fprintf(stream, "\n");
	fprintf(stream, "%20.15e \t del_s\n\n\n",    par.del_s);
// 	fprintf(stream, "%20.15e \t del_t\n\n",  par.del_t);

	print_eos_parameters(par, stream);
	fprintf(stream, "\n");

	for(idom=0; idom<NDOM; idom++){       // Writing the Potential values
		ns=par.ns[idom];		Ns=ns-1;
		nt=par.nt[idom];		Nt=nt-1;
		for(ipot=0; ipot<NPOT; ipot++){
			for(j=0;j<nt;j++) {
				for(i=0;i<ns;i++){
					n_v = Index(par, idom,ipot,i,j);
					n   = par.n_of_n_v[n_v];
					if(n>=0) fprintf(stream, "%20.15e \t (idom,ipot,i,j)=(%d,%d,%d,%d)\n", 
						X[n], idom, ipot, i, j);
				}
			}
		}
	}
	fprintf(stream, "\n");
	nt=par.nt[NDOM-1];
	for(j=0;j<nt;j++) {   // Writing the Surface values
		fprintf(stream, "%20.15e \t Surface_G[j=%d]\n", X[n_2D + j], j);
	}
	fprintf(stream, "\n");   // Writing the Parameter values
	fprintf(stream, "%20.15e \t V0\n",        X[n_2D + nt + 0]);
	fprintf(stream, "%20.15e \t Omega\n",     X[n_2D + nt + 1]);

	fclose(stream);
}
//   -----------------------------------------------
