#include "spheroidal.h"

//  -----------------------------------------------
void func_e2U(parameters par, double *cheb_G, double *cheb_dG, 
	int idom, double s, double t, derivs_2D v, derivs_1D w, derivs_2D V, double *e2U, double *e2U_s, double *e2U_t)
{
	int ipot;
	coordinate xx, yy;

	for(ipot=0; ipot<NPOT; ipot++){
		if(idom == NDOM-1){
			V.d0[ipot]  = interpolate_w(par, ipot, t, w.d0) + interpolate_v(par, idom, ipot, s, t, v.d0);
			V.d1[ipot]  =                                     interpolate_v(par, idom, ipot, s, t, v.d1);
			V.d2[ipot]  = interpolate_w(par, ipot, t, w.d2) + interpolate_v(par, idom, ipot, s, t, v.d2);
		}
		else{
			V.d0[ipot]  = interpolate_v(par, idom, ipot, s, t, v.d0);
			V.d1[ipot]  = interpolate_v(par, idom, ipot, s, t, v.d1);
			V.d2[ipot]  = interpolate_v(par, idom, ipot, s, t, v.d2);
		}
	}
	coord_st_To_xy_ArbitPos(par, idom, cheb_G, cheb_dG, s, t, &xx, &yy);
	
	double
		x    = xx.d0,     x_s    = xx.d1,        x_t    = xx.d2,
		u    = V.d0[0],   u_s    = V.d1[0],      u_t    = V.d2[0],
		B    = V.d0[1],   B_s    = V.d1[1],      B_t    = V.d2[1],
		om   = V.d0[2],   om_s   = V.d1[2],      om_t   = V.d2[2],
	
		B2   = B*B,       B2_s   = 2.*B*B_s,     B2_t   = 2.*B*B_t,
		om2  = om*om,     om2_s  = 2.*om*om_s,   om2_t  = 2.*om*om_t,
		e2u  = exp(2.*u), e2u_s  = 2.*e2u*u_s,   e2u_t  = 2.*e2u*u_t,
		em2u = 1./e2u,    em2u_s = -2.*em2u*u_s, em2u_t = -2.*em2u*u_t;
	
	*e2U   = B2*e2u - x*om2*em2u;
	*e2U_s = (B2_s*e2u + B2*e2u_s) - (x_s*om2*em2u + x*om2_s*em2u + x*om2*em2u_s);
	*e2U_t = (B2_t*e2u + B2*e2u_t) - (x_t*om2*em2u + x*om2_t*em2u + x*om2*em2u_t);
}
//  -----------------------------------------------
double ergo_bisec_t(parameters par, int idom,  double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V,
     double s, double ta, double tb)
{
	int k, bisec_iter = 60;
	double fa, f, fb, a = ta, b = tb, t = 0.5*(a+b), e2U_s, e2U_t;
	
	func_e2U(par, cheb_G, cheb_dG, idom, s, a, v, w, V, &fa, &e2U_s, &e2U_t);
	func_e2U(par, cheb_G, cheb_dG, idom, s, b, v, w, V, &fb, &e2U_s, &e2U_t);
	for(k=0; k<bisec_iter; k++){
		func_e2U(par, cheb_G, cheb_dG, idom, s, t, v, w, V, &f, &e2U_s, &e2U_t);
		if(f*fa > 0) a=t;
		else         b=t;
		t=0.5*(a+b);
	}
	return t;
}	
//  -----------------------------------------------
void ergo_newt_t(parameters par, double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V, double s, double *t)
{
	int idom = 1, k, newt_iter = 5;
	double e2U, e2U_s, e2U_t;
	
	for(k=0; k<newt_iter; k++){
		func_e2U(par, cheb_G, cheb_dG, idom, s, *t, v, w, V, &e2U, &e2U_s, &e2U_t);
		*t -= e2U/e2U_t;
	}
}
//  -----------------------------------------------
void ergo_newt_R(parameters par, int idom,  double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V,
      double t0, double phi, double *R, double *s, double *t)
{
	int k, newt_iter = 5;
	double e2U_s, e2U_t, f, df, sin_phi = sin(phi), cos_phi = cos(phi);
	
	for(k=0; k<newt_iter; k++){
		if(idom == 0) {
			*s = 1. - (*R)*sin_phi;
			*t = 1. - (*R)*cos_phi;
		}
		else{
			*s = (*R)*sin_phi;
			*t = t0 - (*R)*cos_phi;
		}
		func_e2U(par, cheb_G, cheb_dG, idom, *s, *t, v, w, V, &f, &e2U_s, &e2U_t);

		if(idom == 0) df  = -e2U_s*sin_phi - e2U_t*cos_phi;
		else          df  =  e2U_s*sin_phi - e2U_t*cos_phi;

		*R -= f/df;
	}
}
//  -----------------------------------------------
void ergosphere(parameters par, double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V)
{
	int idom, j, bisec_iter = 60, n = 200;
	double s, t, e2U, e2U_s, e2U_t, R, phi;
	coordinate xx, yy;
	FILE *stream;

	stream = fopen("ergosphere.dat", "w");

	func_e2U(par, cheb_G, cheb_dG, 1, 0., 1., v, w, V, &e2U, &e2U_s, &e2U_t);
	if(e2U>0.){ // ergoregion does not extend into the exterior domain 0
		double a=0., b=1., fa, f, fb, t0;
		idom = 1;
		s    = 0.;
		func_e2U(par, cheb_G, cheb_dG, idom, s, a, v, w, V, &e2U, &e2U_s, &fa);
		func_e2U(par, cheb_G, cheb_dG, idom, s, b, v, w, V, &e2U, &e2U_s, &fb);
		if(fa*fb < 0.){ // finding the minimum of e2U in domain 1
			t = 0.5*(a+b);
			for(j=0; j<bisec_iter; j++){
				func_e2U(par, cheb_G, cheb_dG, idom, s, t, v, w, V, &e2U, &e2U_s, &f);
				if(f*fa > 0) a=t;
				else         b=t;
				t=0.5*(a+b);
			}
			if(e2U<0) printf("Star possesses an ergoregion. Please perform `gnuplot surface_and_ergo.plt' \n\n");
			else      printf("Star possesses no ergoregion. Please perform `gnuplot surface.plt' \n\n");
			t0 = t;
			if(e2U < 0.) {// ergoregion exists in domain 1
				t  = ergo_bisec_t(par, idom, cheb_G, cheb_dG, v, w, V, s, 0., t0);
				R  = t0 - t;
				for(j = 0; j<=n; j++){
					phi = Pi*sqr(sin(Pih*j/n));
					ergo_newt_R(par, 1, cheb_G, cheb_dG, v, w, V, t0, phi, &R, &s, &t);
					coord_st_To_xy_ArbitPos(par, idom, cheb_G, cheb_dG, s, t, &xx, &yy);
// 					fprintf(stream, "%e \t %e \n", xx.d0, yy.d0);
					func_e2U(par, cheb_G, cheb_dG, idom, s, t, v, w, V, &e2U, &e2U_s, &e2U_t);
					fprintf(stream, "%e \t %e \t %e \t %d\n", xx.d0, yy.d0, e2U, idom);
				}
			}
		}
		else printf("Star possesses no ergoregion. Please perform `gnuplot surface.plt' \n\n"); // stimmt das?
	}
	else{
		printf("Star possesses an ergoregion. Please perform `gnuplot surface_and_ergo.plt' \n\n");
		idom = 1;
		t  = ergo_bisec_t(par, idom, cheb_G, cheb_dG, v, w, V, 0., 0., 1.);
		for(j = 0; j<=n; j++){
			s = sqr(sin(Pih*j/n));
			ergo_newt_t(par, cheb_G, cheb_dG, v, w, V, s, &t);
			coord_st_To_xy_ArbitPos(par, idom, cheb_G, cheb_dG, s, t, &xx, &yy);
			func_e2U(par, cheb_G, cheb_dG, idom, s, t, v, w, V, &e2U, &e2U_s, &e2U_t);
			fprintf(stream, "%e \t %e \t %e \t %d\n", xx.d0, yy.d0, e2U, idom);
		}
		idom = 0;
		R    = 1. - t;
		for(j = 0; j<=n; j++){
			phi = Pih*sqr(sin(Pih*j/n));
			ergo_newt_R(par, idom, cheb_G, cheb_dG, v, w, V, 1., phi, &R, &s, &t);
			coord_st_To_xy_ArbitPos(par, idom, cheb_G, cheb_dG, s, t, &xx, &yy);
			func_e2U(par, cheb_G, cheb_dG, idom, s, t, v, w, V, &e2U, &e2U_s, &e2U_t);
			fprintf(stream, "%e \t %e \t %e \t %d\n", xx.d0, yy.d0, e2U, idom);
		}
	}
	fclose(stream);
}
//  -----------------------------------------------

