#include "spheroidal.h"

//---------------------------------------------------------------
void pot_st_To_xy(parameters par, coordinate x, coordinate y, derivs_2D V)
{// Calculation of derivatives with respect to x and y from those with respect to s and t
 	double det_inv, 
		d1_V_st[3], d1_V_xy[3], d2_V_st[3][3], d2_V_xy[3][3], 
		d1_xy_st[3][3], d2_xy_st[3][3][3], d1_st_xy[3][3], d2_st_xy[3][3][3];
	int    i, j, k, i1, j1, k1, ipot;
	   
	d1_xy_st[1][1] = x.d1;  // dx/ds
	d1_xy_st[1][2] = x.d2;  // dx/dt
	d1_xy_st[2][1] = y.d1;  // dy/ds
	d1_xy_st[2][2] = y.d2;  // dy/dt

	det_inv = 1./(d1_xy_st[1][1]*d1_xy_st[2][2]-d1_xy_st[1][2]*d1_xy_st[2][1]);
	d1_st_xy[1][1] =  d1_xy_st[2][2]*det_inv;  // ds/dx
	d1_st_xy[1][2] = -d1_xy_st[1][2]*det_inv;  // ds/dy
	d1_st_xy[2][1] = -d1_xy_st[2][1]*det_inv;  // dt/dx
	d1_st_xy[2][2] =  d1_xy_st[1][1]*det_inv;  // dt/dy

	d2_xy_st[1][1][1]                     = x.d11; 	// d^2(x)/(ds)^2
	d2_xy_st[1][1][2] = d2_xy_st[1][2][1] = x.d12;	// d^2(x)/(ds*dt)
	d2_xy_st[1][2][2]                     = x.d22;	// d^2(x)/(dt)^2

	d2_xy_st[2][1][1]                     = y.d11; 	// d^2(y)/(ds)^2
	d2_xy_st[2][1][2] = d2_xy_st[2][2][1] = y.d12;	// d^2(y)/(ds*dt)
	d2_xy_st[2][2][2]                     = y.d22;	// d^2(y)/(dt)^2
	
	for(i=1; i<=2; i++){ // Calculation of d^2(s;t)/[d(x;y)]^2
		for(j=1; j<=2; j++){
			for(k1=1; k1<=2; k1++){
				d2_st_xy[k1][i][j] = 0.;
				for(i1=1; i1<=2; i1++)
					for(j1=1; j1<=2; j1++)
						for(k=1; k<=2; k++)
							d2_st_xy[k1][i][j] -= 
								d2_xy_st[k][i1][j1]*d1_st_xy[i1][i]*d1_st_xy[j1][j]*d1_st_xy[k1][k];
			}
		}
	}
	
	for(ipot=0; ipot < NPOT; ipot++){
		d1_V_st[1]    = V.d1[ipot];
		d1_V_st[2]    = V.d2[ipot];
		d2_V_st[1][1] = V.d11[ipot];
		d2_V_st[1][2] = d2_V_st[2][1] = V.d12[ipot];
		d2_V_st[2][2] = V.d22[ipot];
		for(i=1; i<=2; i++){
			for(j=1; j<=2; j++){
				d2_V_xy[i][j] = 0.;
				for(i1=1; i1<=2; i1++){
					d2_V_xy[i][j] += d1_V_st[i1]*d2_st_xy[i1][i][j];
					for(j1=1; j1<=2; j1++)
						d2_V_xy[i][j] += d2_V_st[i1][j1]*d1_st_xy[i1][i]*d1_st_xy[j1][j];
				}
			}
		}
		V.d11[ipot] = d2_V_xy[1][1];
		V.d12[ipot] = d2_V_xy[1][2];
		V.d22[ipot] = d2_V_xy[2][2];

		d1_V_st[1]  = V.d1[ipot];
		d1_V_st[2]  = V.d2[ipot];
		for(i=1; i<=2; i++){
			d1_V_xy[i] = 0.;
			for(i1=1; i1<=2; i1++)
				d1_V_xy[i] += d1_V_st[i1]*d1_st_xy[i1][i];
		}
		V.d1[ipot]=d1_V_xy[1];
		V.d2[ipot]=d1_V_xy[2];
	}
}
//---------------------------------------------------------------
void D_pot_st_To_xy(parameters par, coordinate x,  coordinate y,   derivs_2D V,
									coordinate Dx, coordinate Dy,  derivs_2D DV)
{// Linear version of 'pot_st_To_xy'
	double det_inv, 
		  d1_V_st[3],   d1_V_xy[3],   d2_V_st[3][3],   d2_V_xy[3][3], 
		D_d1_V_st[3], D_d1_V_xy[3], D_d2_V_st[3][3], D_d2_V_xy[3][3], 
		  d1_xy_st[3][3],   d2_xy_st[3][3][3],   d1_st_xy[3][3],   d2_st_xy[3][3][3],
		D_d1_xy_st[3][3], D_d2_xy_st[3][3][3], D_d1_st_xy[3][3], D_d2_st_xy[3][3][3];
	int i, j, k, i1, j1, k1, ipot;
	   
	d1_xy_st[1][1]   = x.d1;  // dx/ds
	d1_xy_st[1][2]   = x.d2;  // dx/dt
	d1_xy_st[2][1]   = y.d1;  // dy/ds
	d1_xy_st[2][2]   = y.d2;  // dy/dt

	D_d1_xy_st[1][1] = Dx.d1;  // dx/ds
	D_d1_xy_st[1][2] = Dx.d2;  // dx/dt
	D_d1_xy_st[2][1] = Dy.d1;  // dy/ds
	D_d1_xy_st[2][2] = Dy.d2;  // dy/dt
	
	det_inv = 1./(d1_xy_st[1][1]*d1_xy_st[2][2]-d1_xy_st[1][2]*d1_xy_st[2][1]);
	d1_st_xy[1][1] =  d1_xy_st[2][2]*det_inv;  // ds/dx
	d1_st_xy[1][2] = -d1_xy_st[1][2]*det_inv;  // ds/dy
	d1_st_xy[2][1] = -d1_xy_st[2][1]*det_inv;  // dt/dx
	d1_st_xy[2][2] =  d1_xy_st[1][1]*det_inv;  // dt/dy

	for(i=1; i<=2; i++){ 
		for(j=1; j<=2; j++){
			D_d1_st_xy[i][j] = 0.;
			for(i1=1; i1<=2; i1++)
				for(j1=1; j1<=2; j1++)
					D_d1_st_xy[i][j] -= d1_st_xy[i][i1]*D_d1_xy_st[i1][j1]*d1_st_xy[j1][j];
		}
	}
	
	d2_xy_st[1][1][1]                         = x.d11; 	// d^2(x)/(ds)^2
	d2_xy_st[1][1][2]   = d2_xy_st[1][2][1]   = x.d12;	// d^2(x)/(ds*dt)
	d2_xy_st[1][2][2]                         = x.d22;	// d^2(x)/(dt)^2

	d2_xy_st[2][1][1]                         = y.d11; 	// d^2(y)/(ds)^2
	d2_xy_st[2][1][2]   = d2_xy_st[2][2][1]   = y.d12;	// d^2(y)/(ds*dt)
	d2_xy_st[2][2][2]                         = y.d22;	// d^2(y)/(dt)^2
	
	D_d2_xy_st[1][1][1]                       = Dx.d11;
	D_d2_xy_st[1][1][2] = D_d2_xy_st[1][2][1] = Dx.d12;
	D_d2_xy_st[1][2][2]                       = Dx.d22;

	D_d2_xy_st[2][1][1]                       = Dy.d11; 
	D_d2_xy_st[2][1][2] = D_d2_xy_st[2][2][1] = Dy.d12;
	D_d2_xy_st[2][2][2]                       = Dy.d22;
	
	for(i=1; i<=2; i++){ // Calculation of d^2(s;t)/[d(x;y)]^2
		for(j=1; j<=2; j++){
			for(k1=1; k1<=2; k1++){
				d2_st_xy[k1][i][j] = 0.;
				for(i1=1; i1<=2; i1++)
					for(j1=1; j1<=2; j1++)
						for(k=1; k<=2; k++)
							d2_st_xy[k1][i][j] -=
								d2_xy_st[k][i1][j1]*d1_st_xy[i1][i]*d1_st_xy[j1][j]*d1_st_xy[k1][k];
			}
		}
	}
	
	for(i=1; i<=2; i++){ 
		for(j=1; j<=2; j++){
			for(k1=1; k1<=2; k1++){
				D_d2_st_xy[k1][i][j] = 0.;
				for(i1=1; i1<=2; i1++)
					for(j1=1; j1<=2; j1++)
						for(k=1; k<=2; k++)
							D_d2_st_xy[k1][i][j] -=
							   D_d2_xy_st[k][i1][j1]*  d1_st_xy[i1][i]*  d1_st_xy[j1][j]*  d1_st_xy[k1][k]
							+    d2_xy_st[k][i1][j1]*D_d1_st_xy[i1][i]*  d1_st_xy[j1][j]*  d1_st_xy[k1][k]
							+    d2_xy_st[k][i1][j1]*  d1_st_xy[i1][i]*D_d1_st_xy[j1][j]*  d1_st_xy[k1][k]
							+    d2_xy_st[k][i1][j1]*  d1_st_xy[i1][i]*  d1_st_xy[j1][j]*D_d1_st_xy[k1][k];
			}
		}
	}
	
	for(ipot=0; ipot < NPOT; ipot++){
		d1_V_st[1]    = V.d1[ipot];
		d1_V_st[2]    = V.d2[ipot];
		d2_V_st[1][1] = V.d11[ipot];
		d2_V_st[1][2] = d2_V_st[2][1] = V.d12[ipot];
		d2_V_st[2][2] = V.d22[ipot];
		for(i=1; i<=2; i++){
			for(j=1; j<=2; j++){
				d2_V_xy[i][j] = 0.;
				for(i1=1; i1<=2; i1++){
					d2_V_xy[i][j] += d1_V_st[i1]*d2_st_xy[i1][i][j];
					for(j1=1; j1<=2; j1++)
						d2_V_xy[i][j] += 
							d2_V_st[i1][j1]*d1_st_xy[i1][i]*d1_st_xy[j1][j];
				}
			}
		}
		V.d11[ipot] = d2_V_xy[1][1];
		V.d12[ipot] = d2_V_xy[1][2];
		V.d22[ipot] = d2_V_xy[2][2];

		d1_V_st[1]  = V.d1[ipot];
		d1_V_st[2]  = V.d2[ipot];
		for(i=1; i<=2; i++){
			d1_V_xy[i] = 0.;
			for(i1=1; i1<=2; i1++)
				d1_V_xy[i] += d1_V_st[i1]*d1_st_xy[i1][i];
		}
		V.d1[ipot]=d1_V_xy[1];
		V.d2[ipot]=d1_V_xy[2];
	
	
		D_d1_V_st[1]    = DV.d1[ipot];
		D_d1_V_st[2]    = DV.d2[ipot];
		D_d2_V_st[1][1] = DV.d11[ipot];
		D_d2_V_st[1][2] = D_d2_V_st[2][1] = DV.d12[ipot];
		D_d2_V_st[2][2] = DV.d22[ipot];
		for(i=1; i<=2; i++){
			for(j=1; j<=2; j++){
				D_d2_V_xy[i][j] = 0.;
				for(i1=1; i1<=2; i1++){
					D_d2_V_xy[i][j] +=    D_d1_V_st[i1]*  d2_st_xy[i1][i][j]
					                    +   d1_V_st[i1]*D_d2_st_xy[i1][i][j];
					for(j1=1; j1<=2; j1++)
						D_d2_V_xy[i][j] += 
						    D_d2_V_st[i1][j1]*  d1_st_xy[i1][i]*  d1_st_xy[j1][j]
						  +   d2_V_st[i1][j1]*D_d1_st_xy[i1][i]*  d1_st_xy[j1][j]
						  +   d2_V_st[i1][j1]*  d1_st_xy[i1][i]*D_d1_st_xy[j1][j];
				}
			}
		}
		DV.d11[ipot] = D_d2_V_xy[1][1];
		DV.d12[ipot] = D_d2_V_xy[1][2];
		DV.d22[ipot] = D_d2_V_xy[2][2];

		D_d1_V_st[1]  = DV.d1[ipot];
		D_d1_V_st[2]  = DV.d2[ipot];
		for(i=1; i<=2; i++){
			D_d1_V_xy[i] = 0.;
			for(i1=1; i1<=2; i1++)
				D_d1_V_xy[i] +=   D_d1_V_st[i1]*  d1_st_xy[i1][i]
				                 +  d1_V_st[i1]*D_d1_st_xy[i1][i];
		}
		DV.d1[ipot]=D_d1_V_xy[1];
		DV.d2[ipot]=D_d1_V_xy[2];
	}
}
//---------------------------------------------------------------
