#include "spheroidal.h"

void surface(parameters par, double *cheb_G, double *cheb_dG)
{
	double s=1., t;
	int j, n = 200, idom = 1;
	coordinate xx, yy;
	FILE *stream;

	stream = fopen("surface.dat", "w");

	for(j = 0; j<=n; j++){
		t = sqr(sin(Pih*j/n));
		coord_st_To_xy_ArbitPos(par, idom, cheb_G, cheb_dG, s, t, &xx, &yy);
		fprintf(stream, "%e \t %e \n", xx.d0, yy.d0);
	}
	fclose(stream);
}
//  -----------------------------------------------
void output(parameters *par, double *X)
{   // setting and screen output of physical parameters
	int ntotal = (*par).ntotal, n_2D = (*par).n_2D, nt = (*par).nt[NDOM-1], Nt = nt-1;
	derivs_2D v, V;
	derivs_1D G, w;
	double *F, *cheb_G, *cheb_dG;

	allocate_derivs_1D(&G, nt);
	allocate_derivs_1D(&w, nt*NPOT);
	allocate_derivs_2D(&v, n_2D);
	allocate_derivs_2D(&V, NPOT);
	F       = dvector(0, ntotal-1);

	F_of_X(*par, X, v, w, G, par, F);
	GetPhysicalParameters_NonLin(par, v, w, G, V);

	if((*par).printf_phys_param == 1) Print_Parameters(*par);
	
	if((*par).fprintf_phys_param == 1){
		cheb_G  = dvector(0, Nt);
		cheb_dG = dvector(0, Nt);
		Get_cheb_G(*par, G, cheb_G, cheb_dG);

		surface(*par, cheb_G, cheb_dG);
		ergosphere(*par, cheb_G, cheb_dG, v, w, V);
		frintf_PhysQuant_CylCoord(*par, cheb_G, cheb_dG, v, w, V);

		free_dvector(cheb_G,  0, Nt);
		free_dvector(cheb_dG, 0, Nt);
	}

	free_derivs_1D(&G, nt);
	free_derivs_1D(&w, nt*NPOT);
	free_derivs_2D(&v, n_2D);
	free_derivs_2D(&V, NPOT);
	free_dvector(F,       0, ntotal-1);
}
//   -----------------------------------------------
int main() 
{
	char name[50];
	int Newton_iter, i_Seq, n_Seq, ipar, idom, test=0;
	parameters parInit, par, parGoal, parOld;
	double *X0, *X, h;


	printf("\n");
	printf(" *************************************************\n");
	printf(" *        Relativistic Spheroidal Stars          *\n");
	printf(" *************************************************\n\n");

	sprintf(name, "InitialData");	
	ScanFiles(name, &parInit, &X0, &parGoal);
	parInit.printf_phys_param  = 1;
	parInit.fprintf_phys_param = 0;
	printf(" Reading `InitialData' with the following\n\n");
	output(&parInit, X0);


	par = parGoal;
	parOld = parInit;


//	par = parInit;

	#define pI parInit
	#define pG parGoal

	n_Seq = par.n_Seq;
	for(i_Seq = 0; i_Seq <= n_Seq; i_Seq ++){
		printf(" *******************************************************\n");
		printf(" * Calculation of Sequence element No. \t %4d :        *\n", i_Seq);
		printf(" *******************************************************\n");
		if(n_Seq > 0)  h = 1.*i_Seq/n_Seq;
		else           h = 0.;

		if(n_Seq > 0)  par.ns[0] = rint(pI.ns[0] + h*(pG.ns[0] - pI.ns[0]));
		else           par.ns[0] = pG.ns[0];
		par.nt[0] = par.ns[1] = par.nt[1] = par.ns[0];
		par.nst[0] = 0;
		for(idom=0; idom<NDOM; idom++)
			par.nst[idom+1] = par.nst[idom] + par.ns[idom]*par.nt[idom];
		Get_Arrays_n_And_n_v(&par);
		X = dvector(0, par.ntotal-1);
		GetNewResolution(parOld, par, X0, X);
		par.printf_phys_param = 0;
		output(&par, X);
		
		for(ipar=0; ipar<NPAR; ipar++){
			switch(par.choice[ipar]){
				case CHOICE_M        :par.M        = pI.M       + h*(pG.M       - pI.M);       break;
				case CHOICE_M0       :par.M0       = pI.M0      + h*(pG.M0      - pI.M0);      break;
				case CHOICE_J        :par.J        = pI.J       + h*(pG.J       - pI.J);       break;
				case CHOICE_rp       :par.rp       = pI.rp      + h*(pG.rp      - pI.rp);      break;
				case CHOICE_re       :par.re       = pI.re      + h*(pG.re      - pI.re);      break;
				case CHOICE_r_ratio  :par.r_ratio  = pI.r_ratio + h*(pG.r_ratio - pI.r_ratio); break;
				case CHOICE_beta     :par.beta     = pI.beta    + h*(pG.beta    - pI.beta);    break;
				case CHOICE_V0       :par.V0       = pI.V0      + h*(pG.V0      - pI.V0);      break;
				case CHOICE_Omega    :par.Omega    = pI.Omega   + h*(pG.Omega   - pI.Omega);   break;
				case CHOICE_p_c      :par.p_c      = pI.p_c     + h*(pG.p_c     - pI.p_c);     break;
				case CHOICE_eps_c    :par.eps_c    = pI.eps_c   + h*(pG.eps_c   - pI.eps_c);   break;
				case CHOICE_muB_c    :par.muB_c    = pI.muB_c   + h*(pG.muB_c   - pI.muB_c);   break;
				case CHOICE_h_max    :par.h_max    = pI.h_max   + h*(pG.h_max   - pI.h_max);   break;
				case CHOICE_B_c      :par.B_c      = pI.B_c     + h*(pG.B_c     - pI.B_c);     break;
				case CHOICE_Eb       :par.Eb       = pI.Eb      + h*(pG.Eb      - pI.Eb);      break;
				case CHOICE_Rcirc    :par.Rcirc    = pI.Rcirc   + h*(pG.Rcirc   - pI.Rcirc);   break;
			}
		}
		par.del_s   = pI.del_s   + h*(pG.del_s-pI.del_s);
		par.del_t   = pI.del_t   + h*(pG.del_t-pI.del_t);
		par.h0      = pI.h0      + h*(pG.h0 - pI.h0);
		switch(par.eos){
			case 0: // for homogeneous body
				{
					par.eps0  = pI.eps0 + h*(pG.eps0 - pI.eps0);
					break;
				}
			case 1: // for polytropic body
				{
					par.N_poly  = pI.N_poly + h*(pG.N_poly - pI.N_poly);
					par.K_poly  = pI.K_poly + h*(pG.K_poly - pI.K_poly);
					break;
				}
			case 2: // for ideal, relativistic neutron gas,
					// normalization: for p->0 one finds the polytrope with n=3/2
				{
					break;
				}
			case 3: // customized equation of state
				{
					int k, Np = par.eos3_Np;
					par.h0 = pI.h0 + h*(pG.h0 - pI.h0);
					for(k=1; k <= Np; k++)
						par.eos3_p[k] = pI.eos3_p[k] + h*(pG.eos3_p[k] - pI.eos3_p[k]);
				}
		}

// 		newton_old(par, X);
		Newton_iter = newton(par, X);

		printf("\n Writing File `%s.", par.SeqName);
		if (i_Seq <  10)              printf("00%1d", i_Seq);
		if (i_Seq >= 10 && i_Seq<100) printf( "0%2d", i_Seq);
		if (i_Seq >= 100)             printf(  "%3d", i_Seq);
		printf("' with the following\n\n");
		PrintToFile(par, X, i_Seq);

		par.printf_phys_param = 1;
		if (i_Seq < n_Seq) par.fprintf_phys_param = 0;
		else               par.fprintf_phys_param = 1;
		output(&par, X);
		parOld = par;

		free_dvector(X0, 0, par.ntotal-1);
		X0 = X;
	}

	#undef pI
	#undef pG

	free_dvector(X,  0, parGoal.ntotal-1);
//	Free_Arrays_n_And_n_v(&parGoal);
	Free_Arrays_n_And_n_v(&par);

	return 1;
}
