#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "utilities.h"

#define NDOM 				2
#define NPOT 				4
#define NPAR 				2
#define NpMAX 				20

#define CHOICE_M          1     //   1  M
#define CHOICE_M0         2     //   2  M0
#define CHOICE_J          3     //   3  J
#define CHOICE_rp         4     //   4  rp
#define CHOICE_re         5     //   5  re
#define CHOICE_r_ratio    6     //   6  r_ratio
#define CHOICE_beta       7     //   7  beta
#define CHOICE_V0         8     //   8  V0
#define CHOICE_Omega      9     //   9  Omega
#define CHOICE_p_c       10     //  10  p_c
#define CHOICE_eps_c     11     //  11  eps_c
#define CHOICE_muB_c     12     //  12  muB_c
#define CHOICE_h_max     13     //  13  h_max
#define CHOICE_B_c       14     //  14  B_c
#define CHOICE_Eb        15     //  15  Eb
#define CHOICE_Rcirc     16     //  16  Rcirc

typedef struct PARAMETERS{
	int
		eos, // Specification of equation of state (see 'eqstate.c')
		eos3_Np,
		// in X[0]..X[num_v] all non-trivial potentials v in domaint 0 and 1 are stored
		// [in domain 1, v has the trivial values v(s=1,t) = 0]
		ntotal, num_v, n_2D,
		ns[NDOM], nt[NDOM], nst[NDOM+1], *n_of_n_v , *n_v_of_n, choice[NPAR], 
		n_Seq,  // Number_of_Sequence_Solutions
		nx, ny, // Number of cylindrical Grid points for the file "PhysQuant_CylCoord.dat"
		printf_phys_param, fprintf_phys_param,
		Newton_itmin, Newton_itmax, Newton_verb,
		bicgstab_itmax, bicgstab_verb,
		PreCond_itmax, PreCond_FD_Order,
		test_wr_Jac;
	double	del_s, del_t,
		eps0,              // eps0 for homogeneous stars
		K_poly, N_poly,    // K and N for polytropic stars
		a_strange, B_strange, muB0_strange,       // a, B and muB0 for strange stars
		eos3_muB0, eos3_p[NpMAX],
		M, M0, Mp, W, T, J, Eb,
		rp, re, xp, xe, r_ratio, Rcirc, beta, V0, Omega,
		h0, h_c, p_c, eps_c, muB_c, u_c, B_c, om_c, mu_c, h_max, t_h_max,
		M_in, J_in, M_test, J_test,
		xmax, ymax, // Cylindrical borders for the file "PhysQuant_CylCoord.dat"
		Newton_tol,
		bicgstab_decr, bicgstab_tol,
		PreCond_Improve;
	char SeqName[200];
} parameters;

typedef struct COORDINATE{ 
	double d0, d1, d2, d11, d12, d22, d122;
} coordinate;

typedef struct DERIVS_1D{
	double *d0, *d2, *d22;
} derivs_1D;


typedef struct DERIVS_2D{ 
	double *d0, *d1, *d2, *d11, *d12, *d22;
} derivs_2D;

typedef struct JFD_COMPONENTS{ 
	double 
		**J_UL, **K_UL, **Kl_UL,
		**J_UR, **K_UR,
		**J_LL,
		**J_LR, **K_LR;
	int
		*ncols_J_UL, **cols_J_UL, *iK_UL,
		*iK_LR,
		m1, m2;
} JFD_Components;


// Subroutines in "FlatStar.c"
void surface(parameters par, double *cheb_G, double *cheb_dG);
void output(parameters *par, double *X);
int main();

// Subroutines in "pot_st_To_xy.c"
void pot_st_To_xy(parameters par, coordinate x, coordinate y, derivs_2D V);
void D_pot_st_To_xy(parameters par, coordinate x,  coordinate y,   derivs_2D V,
									coordinate Dx, coordinate Dy,  derivs_2D DV);

// Subroutines in "CoordTransf.c"
void coord_st_To_xy(parameters par, int idom, int i, int j, derivs_1D G, coordinate *x, coordinate *y);
void Dcoord_st_To_xy(parameters par, parameters Dpar, int idom, int i, int j, derivs_1D G, derivs_1D DG, 
	coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy);
void get_sig(parameters par, int idom, int i, double *sig, double *sig_m_1, double *sig_s, double *sig_ss);
void get_tau(parameters par, int idom, int j, double *tau, double *tau_m_1, double *tau_t, double *tau_tt);
void coord_0(parameters par, int i, int j, derivs_1D G, coordinate *x, coordinate *y);
void Dcoord_0(parameters par, parameters Dpar, int i, int j, derivs_1D G, derivs_1D DG, 
	coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy);
void coord_1(parameters par, int i, int j, derivs_1D G, coordinate *x, coordinate *y);
void Dcoord_1(parameters par, parameters Dpar, int i, int j, derivs_1D G, derivs_1D DG, 
	coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy);

// Subroutines in "FuncAndJacobian.c"
void v_To_V(parameters par, int idom, int i, int j, derivs_2D v, derivs_1D w, derivs_2D V);
void Derivatives_st(parameters par, derivs_2D v);
void Derivatives_t(parameters par, derivs_1D w, derivs_1D G);
void Get_vwGpar_From_X(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters *parX);
void F_of_X(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters *parX, double *F);
void J_times_DX(parameters par, double *X, double *DX, derivs_2D v, derivs_1D w, derivs_1D G,
        parameters parX, double *JDX);

// Subroutines in "IndexRoutines.c"
int GradIndex(int i, int j);
int Index_i(parameters par, int idom, int i);
int Index_j(parameters par, int idom, int j);
int Index(parameters par, int idom, int ipot, int i, int j);
int Index_Star(parameters par, int ipot, int j);
void Get_Arrays_n_And_n_v_domain_0(parameters *par, int *n);
void Get_Arrays_n_And_n_v_domain_1(parameters *par, int *n);
void Get_Arrays_n_And_n_v(parameters *par);
void Get_Indices_From_n_v(parameters par, int n_v, int *idom, int *ipot, int *i, int *j);
void Free_Arrays_n_And_n_v(parameters *par);

// Subroutines in "FieldAndBoundEqns.c"
void NonLinPotentials(parameters par, int idom, int i, int j, coordinate xx, coordinate yy, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *Laplace, double *grad_grad);
void LinPotentials(parameters par, int idom, int i, int j, coordinate  xx,  coordinate yy, 
	derivs_2D v,  derivs_1D w,  derivs_1D G,  derivs_2D V, double *Laplace,  double *grad_grad,
	coordinate Dxx, coordinate Dyy, derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV,
	double *DLaplace, double *Dgrad_grad);
void NonLinFieldEqns(parameters par, int idom, int i, int j, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F);
void LinFieldEqns(parameters par, parameters Dpar, int idom, int i, int j,
    derivs_2D  v, derivs_1D  w, derivs_1D  G, derivs_2D  V,
    derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX);
void NonLinStarBoundary(parameters par, int j, derivs_1D w, derivs_1D G, derivs_2D V, double *F);
void LinStarBoundary(parameters par, parameters Dpar, int j, derivs_1D w, derivs_1D G, derivs_2D V,
	derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX);
void NonLinInfinityEqns(parameters par, int idom, int i, int j, derivs_2D v, double *F);
void LinInfinityEqns(parameters par, int idom, int i, int j, derivs_2D Dv, double *JDX);
void NonLinTransitEqns(parameters par, int idom, int idom1, int idom2, int is, int i, int j, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F);
void LinTransitEqns(parameters par, parameters Dpar, int idom, int idom1, int idom2, int is, int i, int j, 
	derivs_2D  v, derivs_1D  w, derivs_1D  G, derivs_2D  V,
	derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double* JDX);
void NonLinBoundEqns(parameters par, int idom, int i, int j, 
						derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F);
void LinBoundEqns(parameters par, parameters Dpar, int idom, int i, int j, 
					derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V,
					derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX);

// Subroutines in "ScanAndPrint.c"
void ScanInitialData(char *name, parameters *parInit, double **X);
void ScanConfig(parameters *parInit, parameters *parGoal);
void ScanFiles(char *name, parameters *parInit, double **X, parameters *parGoal);
void PrintToFile(parameters par, double *X, int i_Seq);

// Subroutines in "GetNewResolution.c"
void v1_To_v2(parameters par1, derivs_2D v1, parameters par2, derivs_2D v2);
void wG1_To_wG2(parameters par1, derivs_1D w1, derivs_1D G1, parameters par2, derivs_1D w2, derivs_1D G2); 
void GetNewResolution(parameters par0, parameters par, double *X0, double *X);

// Subroutines in "allocate_and_free_derivs.c"
void allocate_derivs_1D(derivs_1D *v, int n);
void allocate_derivs_2D(derivs_2D *v, int n);
void fill0_derivs_1D(derivs_1D v, int n);
void fill0_derivs_2D(derivs_2D v, int n);
void free_derivs_1D(derivs_1D *v, int n);
void free_derivs_2D(derivs_2D *v, int n);

// Subroutines in "eqstate.c"
void scan_eos_parameters( parameters *par, FILE *stream);
void print_eos_parameters(parameters  par, FILE *stream);
void display_eos_parameters(parameters par);
double eos3_pressure(parameters par, double H);
double eos3_dpressure(parameters par, double H);
double eos3_d2pressure(parameters par, double H);
void eos_quantities(parameters par, double h, double *p, double *eps, double *muB);
void D_eos_quantities(parameters par, double h, double Dh, double *Dp, double *Deps, double *DmuB);

// Subroutines in "ParameterEqns.c"
void h_max(parameters *par, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V);
void D_hmax(parameters *par, parameters *Dpar,
        derivs_2D v,   derivs_1D w,  derivs_1D G, derivs_2D V,
        derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV);
void M_and_J_in(parameters *par, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V);
void D_M_and_J_in(parameters *par, parameters *Dpar,
       derivs_2D v,   derivs_1D w,  derivs_1D G, derivs_2D V,
       derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV);
void M_and_J_out(parameters *par, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, int idom, int i);
void D_M_and_J(parameters *par, parameters *Dpar,
       derivs_2D v,   derivs_1D w,  derivs_1D G, derivs_2D V, derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, 
	   int idom, int i);
double ShedParameter(parameters par, derivs_1D G);
double D_ShedParameter(parameters par, parameters Dpar, derivs_1D G, derivs_1D DG);
void CenterParameters(parameters *par, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V);
void D_CenterParameters(parameters *par, parameters *Dpar,
       derivs_2D v,   derivs_1D w,  derivs_1D G,  derivs_2D V,
       derivs_2D Dv, derivs_1D Dw, derivs_1D DG,  derivs_2D DV);
void Rcirc(parameters *par, derivs_2D v, derivs_1D w, derivs_2D V);
void D_Rcirc(parameters *par, parameters *Dpar, 
	derivs_2D v, derivs_1D w, derivs_2D V, derivs_2D Dv, derivs_1D Dw, derivs_2D DV);
void GetPhysicalParameters_NonLin(parameters *parX, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V);
void GetPhysicalParameters_Lin(parameters *parX, parameters *DparX,
       derivs_2D v,   derivs_1D w,  derivs_1D G,  derivs_2D V,
       derivs_2D Dv, derivs_1D Dw, derivs_1D DG,  derivs_2D DV);
void NonLinParamEqns(parameters par, parameters parX, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F);
void LinParamEqns(parameters par, parameters parX, parameters DparX, 
					derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V,
					derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX);
void Print_Parameters(parameters par);

// Subroutines in "newton.c"
void resid(parameters par, double *DX, double *rhs, JFD_Components JFD, double *res);
void Solve_JFD_UL(parameters par, double *F, double *X, JFD_Components JFD);
void Solve_JFD(parameters par, double *F, double *DX, JFD_Components JFD);
int PreCond(parameters par, double *F, double *DX, double *norm, JFD_Components JFD);
int bicgstab_Iterations(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters parX, 
		double *F, double *DX, double *normres, JFD_Components JFD);
int bicgstab(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters parX, double *F, double *DX,
     double *normres);
int newton(parameters par, double *X);

// Subroutines in "JacobianFD.c"
void GetFD_Dv_a(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j);
void GetFD_Dv_b(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j);
void GetFD_Dv_ab_FD2(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j);
void GetFD_Dv_ab_FD4(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j);
void GetFD_Dv_st(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j);
void GetFD_DG_t(parameters par, derivs_1D DG, int j);
void GetFD_Dw_t(parameters par, derivs_1D Dw, int ipot, int j);
void GetFD_DvwGpar(parameters par, int n, double *DX, derivs_2D v, derivs_1D w, derivs_1D G, 
	derivs_2D Dv, derivs_1D Dw, derivs_1D DG, parameters *DparX);
void fill0_Dv_of_indx(int indx, derivs_2D Dv);
void fill0_DvwGpar(parameters par, int n, derivs_2D Dv, derivs_1D Dw, derivs_1D DG, parameters *DparX);
void Get_boundary_i1(parameters par, int jdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1,
		int *mdom);
void Get_gridpoints(parameters par, int n, int *mdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1);
void Get_JFD_Matrix(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G,
          parameters parX, JFD_Components *JFD);
void Get_JFD_Components(parameters par, JFD_Components *JFD);

// Subroutines in "InterpolationRoutines.c.c"
double s_of_sig(parameters par, double sig);
double t_of_tau(parameters par, double tau);
double s_of_xi2(parameters par, double xi2);
void get_sig_ArbitPos(parameters par, double s, double *sig, double *sig_m_1, double *sig_s);
void get_tau_ArbitPos(parameters par, double t, double *tau, double *tau_m_1, double *tau_t);
void coord_st_To_xy_ArbitPos(parameters par, int idom, double *cheb_G, double *cheb_dG, double s, double t, 
	coordinate *x, coordinate *y);
void coord_0_ArbitPos(parameters par, double *cheb_G, double *cheb_dG, double s, double t, coordinate *x, coordinate *y);
void coord_1_ArbitPos(parameters par, double *cheb_G, double *cheb_dG, double s, double t, coordinate *x, coordinate *y);
void xy_To_st(parameters par, double *cheb_G, double *cheb_dG,	double x, double y, int *idom, double *s, double *t);
double interpolate_w(parameters par, int ipot, double t, double *w);
double interpolate_v(parameters par, int idom, int ipot, double s, double t, double *v);
void v_To_V_ArbitPos(parameters par, double *cheb_G, double *cheb_dG, 
	int idom, double s, double t, derivs_2D v, derivs_1D w, derivs_2D V);
void Get_cheb_G(parameters par, derivs_1D G, double *cheb_G, double *cheb_dG);
void Interpolation_To_Cylindrical_Gridpoint(parameters par, double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V,
	double x, double y, int *idom, double *s, double *t, double *h, double *p, double *eps, double *muB);
void frintf_PhysQuant_CylCoord(parameters par, double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V);

// Subroutines in "ergosphere.dat"
void func_e2U(parameters par, double *cheb_G, double *cheb_dG, 
	int idom, double s, double t, derivs_2D v, derivs_1D w, derivs_2D V, double *e2U, double *e2U_s, double *e2U_t);
double ergo_bisec_t(parameters par, int idom,  double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V,
     double s, double ta, double tb);
void ergo_newt_t(parameters par, double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V, double s, double *t);
void ergo_newt_R(parameters par, int idom,  double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V,
      double t0, double phi, double *R, double *s, double *t);
void ergosphere(parameters par, double *cheb_G, double *cheb_dG, derivs_2D v, derivs_1D w, derivs_2D V);


// Subroutines in "newton_old.c"
// void Jacobian(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, double **DF);
// void DF_of_X(   parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, double **DF);
// void newton_old(parameters par, double *X);

