#define Pi       3.14159265358979323846264338328
#define Pih      1.57079632679489661923132169164 // Pi/2
#define Piq      0.78539816339744830961566084582 // Pi/4
#define Third    0.33333333333333333333333333333 // 1/3
#define TwoThird 0.66666666666666666666666666667 // 2/3

#define TINY 1.0e-20
#define BIG  1.0e+10
#define SWAP(a,b) {temp=(a);(a)=(b);(b)=temp;}
#define NR_END 1
#define FREE_ARG char*
#define NMAX 100

typedef struct DCOMPLEX {double r,i;} dcomplex;

void nrerror(char error_text[]);
int      *ivector( int nl,  int nh);
double   *dvector( int nl,  int nh);
double  **dpvector(int nl, int nh);
int     **imatrix( int nrl, int nrh, int ncl, int nch);
double  **dmatrix( int nrl, int nrh, int ncl, int nch);
double ***d3tensor(int nrl, int nrh, int ncl, int nch, int ndl, int ndh);
void free_ivector(int       *v, int nl,  int nh);
void free_dvector(double    *v, int nl,  int nh);
void free_dpvector(double   **v, int nl,  int nh);
void free_imatrix(int      **m, int nrl, int nrh, int ncl, int nch);
void free_dmatrix(double   **m, int nrl, int nrh, int ncl, int nch);
void free_d3tensor(double ***t, int nrl, int nrh, int ncl, int nch,
								int ndl, int ndh);
void fill0_dvector(double *X, int n0, int ntotal);
void fill0_ivector(int *X, int n0, int ntotal);
void fill0_dmatrix(double **X, int m0, int mtotal, int n0, int ntotal);
void fill0_imatrix(int **X, int m0, int mtotal, int n0, int ntotal);
void copy_dvector(double *aout, double *ain, int n0, int ntotal);
double norm1(double *v, int n);
double norm2(double *v, int n);
double scalarproduct(double *v, double *w, int n);

int minimum2(int i,int j);
int minimum3(int i,int j,int k);
int maximum2(int i,int j);
double dmaximum2(double a, double b);
int maximum3(int i,int j,int k);
int pow_int(int mantisse,int exponent);
double sinch(double x);
double Sqrt(double x);
double sqr(double x);

dcomplex Cadd(dcomplex a, dcomplex b);
dcomplex Csub(dcomplex a, dcomplex b);
dcomplex Cmul(dcomplex a, dcomplex b);
dcomplex RCmul(double x, dcomplex a);
dcomplex Cdiv(dcomplex a, dcomplex b);
dcomplex Complex(double re, double im);
dcomplex Conjg(dcomplex z);
double Cabs(dcomplex z);

dcomplex Csqrt(dcomplex z);
dcomplex Cexp(dcomplex z);
dcomplex Clog(dcomplex z);
dcomplex Csin(dcomplex z);
dcomplex Ccos(dcomplex z);
dcomplex Ctan(dcomplex z);
dcomplex Ccot(dcomplex z);
dcomplex Csinh(dcomplex z);
dcomplex Ccosh(dcomplex z);
dcomplex Ctanh(dcomplex z);
dcomplex Ccoth(dcomplex z);

void chebft_Zeros(double u[], int n, int inv);
void chebft_Extremes(double u[], int n, int inv);
void chebft_Extremes_2D(double** u, int ns, int nt);
void chder(double a, double b, double c[], double cder[], int n);
double chebev(double a, double b, double c[], int m, double x);
void chint(double a, double b, double c[], double cint[], int n);
void fourft(double *u, int N, int inv);
void fourder(double u[], double du[], int N);
void fourder2(double u[], double d2u[], int N);
double fourev(double *u, int N, double x);


void ludcmp(double **a, int n, int *indx, double *d);
void ludcmp_1(double **a, int n, int *indx, double *d);
void lubksb(double **a, int n, int *indx, double b[]);
void lubksb_1(double **a, int n, int *indx, double b[]);
void bandec(double **a, int n, int m1, int m2, double **al, int indx[], double *d);
void banbks(double **a, int n, int m1, int m2, double **al, int indx[], double b[]);



